"""

postprocess_batch_results.py

Given a .json or .csv file containing MD results, do one or more of the following:

* Sample detections/non-detections and render to HTML (when ground truth isn't
  available) (this is 99.9% of what this module is for)
* Evaluate detector precision/recall, optionally rendering results (requires
  ground truth)
* Sample true/false positives/negatives and render to HTML (requires ground
  truth)

Ground truth, if available, must be in COCO Camera Traps format:

https://github.com/agentmorris/MegaDetector/blob/main/megadetector/data_management/README.md#coco-camera-traps-format

"""

#%% Constants and imports

import argparse
import collections
import copy
import errno
import io
import os
import sys
import time
import uuid
import warnings
import random

from enum import IntEnum
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from functools import partial
from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
import humanfriendly
import pandas as pd

from sklearn.metrics import precision_recall_curve, confusion_matrix, average_precision_score
from tqdm import tqdm

from megadetector.visualization import visualization_utils as vis_utils
from megadetector.visualization import plot_utils
from megadetector.utils.write_html_image_list import write_html_image_list
from megadetector.utils.wi_taxonomy_utils import load_md_or_speciesnet_file
from megadetector.utils import path_utils
from megadetector.utils.ct_utils import args_to_object
from megadetector.utils.ct_utils import sets_overlap
from megadetector.utils.ct_utils import sort_dictionary_by_value
from megadetector.utils.ct_utils import sort_dictionary_by_key
from megadetector.utils.ct_utils import invert_dictionary
from megadetector.utils.path_utils import clean_filename
from megadetector.data_management.cct_json_utils import CameraTrapJsonUtils
from megadetector.data_management.cct_json_utils import IndexedJsonDb
from megadetector.postprocessing.load_api_results import load_api_results
from megadetector.detection.run_detector import get_typical_confidence_threshold_from_results

warnings.filterwarnings('ignore', '(Possibly )?corrupt EXIF data', UserWarning)


#%% Options

DEFAULT_NEGATIVE_CLASSES = ['empty']
DEFAULT_UNKNOWN_CLASSES = ['unknown', 'unlabeled', 'ambiguous']

# Make sure there is no overlap between the two sets, because this will cause
# issues in the code
assert not sets_overlap(DEFAULT_NEGATIVE_CLASSES, DEFAULT_UNKNOWN_CLASSES), (
        'Default negative and unknown classes cannot overlap.')


class PostProcessingOptions:
    """
    Options used to parameterize process_batch_results().
    """

    def __init__(self):

        ### Required inputs

        #: MD results .json file to process
        self.md_results_file = ''

        #: Folder to which we should write HTML output
        self.output_dir = ''

        ### Options

        #: Folder where images live (filenames in [md_results_file] should be relative to this folder)
        #:
        #: Can be '' if [md_results_file] uses absolute paths.
        self.image_base_dir = ''

        ## These apply only when we're doing ground-truth comparisons

        #: Optional .json file containing ground truth information
        self.ground_truth_json_file = ''

        #: List of classes we'll treat as negative (defaults to "empty", typically includes
        #: classes like "blank", "misfire", etc.).
        #:
        #: Include the token "#NO_LABELS#" to indicate that an image with no annotations
        #: should be considered empty.
        self.negative_classes = DEFAULT_NEGATIVE_CLASSES

        #: List of classes we'll treat as neither positive nor negative (defaults to
        #: "unknown", typically includes classes like "unidentifiable").
        self.unlabeled_classes = DEFAULT_UNKNOWN_CLASSES

        #: List of output sets that we should count, but not render images for.
        #:
        #: Typically used to preview sets with lots of empties, where you don't want to
        #: subset but also don't want to render 100,000 empty images.
        #:
        #: Example strings that are valid for this option:
        #:
        #: detections, non_detections
        #: detections_animal, detections_person, detections_vehicle
        #: tn, tp, fn, fp
        self.rendering_bypass_sets = []

        #: If this is None, a confidence threshold is selected based on the detector version.
        #:
        #: This can either be a float or a dictionary mapping category names (not IDs) to
        #: thresholds.  The category "default" can be used to specify thresholds for
        #: other categories.  Currently the use of a dict here is not supported when
        #: ground truth is supplied.
        self.confidence_threshold = None

        #: Confidence threshold to apply to classification (not detection) results
        #:
        #: Only a float is supported here (unlike the "confidence_threshold" parameter, which
        #: can be a dict).
        self.classification_confidence_threshold = 0.5

        #: Used for summary statistics only
        self.target_recall = 0.9

        #: Number of images to sample, -1 for "all images"
        self.num_images_to_sample = 500

        #: Random seed for sampling, or None
        self.sample_seed = 0 # None

        #: Image width for images in the HTML output
        self.viz_target_width = 800

        #: Line width (in pixels) for rendering detections
        self.line_thickness = 4

        #: Box expansion (in pixels) for rendering detections
        self.box_expansion = 0

        #: Job name to include in big letters in the output HTML
        self.job_name_string = None

        #: Model version string to include in the output HTML
        self.model_version_string = None

        #: Sort order for the output, should be one of "filename", "confidence", or "random"
        self.html_sort_order = 'filename'

        #: If True, images in the output HTML will be links back to the original images
        self.link_images_to_originals = True

        #: Optionally separate detections into categories (animal/vehicle/human)
        #:
        #: Currently only supported when ground truth is unavailable
        self.separate_detections_by_category = True

        #: Optionally replace one or more strings in filenames with other strings;
        #: useful for taking a set of results generated for one folder structure
        #: and applying them to a slightly different folder structure.
        self.api_output_filename_replacements = {}

        #: Optionally replace one or more strings in filenames with other strings;
        #: useful for taking a set of results generated for one folder structure
        #: and applying them to a slightly different folder structure.
        self.ground_truth_filename_replacements = {}

        #: Allow bypassing API output loading when operating on previously-loaded
        #: results.  If present, this is a Pandas DataFrame.  Almost never useful.
        self.api_detection_results = None

        #: Allow bypassing API output loading when operating on previously-loaded
        #: results.  If present, this is a str --> obj dict.  Almost never useful.
        self.api_other_fields = None

        #: Should we also split out a separate report about the detections that were
        #: just below our main confidence threshold?
        #:
        #: Currently only supported when ground truth is unavailable.
        self.include_almost_detections = False

        #: Only a float is supported here (unlike the "confidence_threshold" parameter, which
        #: can be a dict).
        self.almost_detection_confidence_threshold = None

        #: Enable/disable rendering parallelization
        self.parallelize_rendering = False

        #: Number of threads/processes to use for rendering parallelization
        self.parallelize_rendering_n_cores = 16

        #: Whether to use threads (True) or processes (False) for rendering parallelization
        self.parallelize_rendering_with_threads = True

        #: When classification results are present, should be sort alphabetically by class
        #: name (False) or in descending order by frequency (True)?
        self.sort_classification_results_by_count = False

        #: When classification results are present, use this dictionary to push some
        #: categories to the bottom of the list.  Larger numbers == later groups.
        #: Default sort weight is zero.  Line breaks will separate equal sort weights.
        #: Sort weights must be integers.
        #:
        #: In practice this is used to push generic categories like "blank", "animal",
        #: and "unreliable" to the bottom of the list, like:
        #:
        #: options.category_name_to_sort_weight = \
        #:     {'animal':1,'blank':1,'unknown':1,'unreliable':1,'mammal':1,'no cv result':1}
        self.category_name_to_sort_weight = {}

        #: Should we split individual pages up into smaller pages if there are more than
        #: N images?
        self.max_figures_per_html_file = None

        #: Footer text for the index page
        # self.footer_text = \
        #    '<br/><p style="font-size:80%;">Preview page created with the ' + \
        #    <a href="{}">MegaDetector Python package</a>.</p>'.\
        #    format('https://megadetector.readthedocs.io')
        self.footer_text = ''

        #: Character encoding to use when writing the index HTML html
        self.output_html_encoding = None

        #: Additional image fields to display in image headers.  If this is a list,
        #: we'll include those fields; if this is a dict, we'll use that dict to choose
        #: alternative display names for each field.
        self.additional_image_fields_to_display = None

        #: If classification results are present, should we include a summary of
        #: classification categories?
        self.include_classification_category_report = True

        #: The category/count summary typically only includes category names, this
        #: flag includes descriptions (typically taxonomic strings) as well.
        self.include_category_descriptions_with_global_counts = False

    # ...__init__()

# ...PostProcessingOptions


class PostProcessingResults:
    """
    Return format from process_batch_results
    """

    def __init__(self):

        #: HTML file to which preview information was written
        self.output_html_file = ''

        #: Pandas Dataframe containing detection results
        self.api_detection_results = None

        #: str --> obj dictionary containing other information loaded from the results file
        self.api_other_fields = None


##%% Helper classes and functions

class DetectionStatus(IntEnum):
    """
    Flags used to mark images as positive or negative for P/R analysis
    (according to ground truth and/or detector output)

    :meta private:
    """

    DS_NEGATIVE = 0
    DS_POSITIVE = 1

    # Anything greater than this isn't clearly positive or negative
    DS_MAX_DEFINITIVE_VALUE = DS_POSITIVE

    # image has annotations suggesting both negative and positive
    DS_AMBIGUOUS = 2

    # image is not annotated or is annotated with 'unknown', 'unlabeled', ETC.
    DS_UNKNOWN = 3

    # image has not yet been assigned a state
    DS_UNASSIGNED = 4

    # In some analyses, we add an additional class that lets us look at
    # detections just below our main confidence threshold
    DS_ALMOST = 5


def _mark_detection_status(indexed_db,
                           negative_classes=DEFAULT_NEGATIVE_CLASSES,
                           unknown_classes=DEFAULT_UNKNOWN_CLASSES):
    """
    For each image in indexed_db.db['images'], add a '_detection_status' field
    to indicate whether to treat this image as positive, negative, ambiguous,
    or unknown.

    Makes modifications in-place.

    returns (n_negative, n_positive, n_unknown, n_ambiguous)
    """

    negative_classes = set(negative_classes)
    unknown_classes = set(unknown_classes)

    # count the # of images with each type of DetectionStatus
    n_unknown = 0
    n_ambiguous = 0
    n_positive = 0
    n_negative = 0

    print('Preparing ground-truth annotations')
    for im in tqdm(indexed_db.db['images']):

        image_id = im['id']
        annotations = indexed_db.image_id_to_annotations[image_id]
        categories = [ann['category_id'] for ann in annotations]
        category_names = set(indexed_db.cat_id_to_name[cat] for cat in categories)

        # Check whether this image has:
        # - unknown / unassigned-type labels
        # - negative-type labels
        # - positive labels (i.e., labels that are neither unknown nor negative)
        has_unknown_labels = sets_overlap(category_names, unknown_classes)
        has_negative_labels = sets_overlap(category_names, negative_classes)
        has_positive_labels = 0 < len(category_names - (unknown_classes | negative_classes))
        # assert has_unknown_labels is False, '{} has unknown labels'.format(annotations)

        # If there are no image annotations...
        if len(categories) == 0:

            if '#NO_LABELS#' in negative_classes:
                n_negative += 1
                im['_detection_status'] = DetectionStatus.DS_NEGATIVE
            else:
                n_unknown += 1
                im['_detection_status'] = DetectionStatus.DS_UNKNOWN

            # n_negative += 1
            # im['_detection_status'] = DetectionStatus.DS_NEGATIVE

        # If the image has more than one type of labels, it's ambiguous
        # note: bools are automatically converted to 0/1, so we can sum
        elif (has_unknown_labels + has_negative_labels + has_positive_labels) > 1:
            n_ambiguous += 1
            im['_detection_status'] = DetectionStatus.DS_AMBIGUOUS

        # After the check above, we can be sure it's only one of positive,
        # negative, or unknown.
        #
        # Important: do not merge the following 'unknown' branch with the first
        # 'unknown' branch above, where we tested 'if len(categories) == 0'
        #
        # If the image has only unknown labels
        elif has_unknown_labels:
            n_unknown += 1
            im['_detection_status'] = DetectionStatus.DS_UNKNOWN

        # If the image has only negative labels
        elif has_negative_labels:
            n_negative += 1
            im['_detection_status'] = DetectionStatus.DS_NEGATIVE

        # If the images has only positive labels
        elif has_positive_labels:
            n_positive += 1
            im['_detection_status'] = DetectionStatus.DS_POSITIVE

            # Annotate the category, if it is unambiguous
            if len(category_names) == 1:
                im['_unambiguous_category'] = list(category_names)[0]

        else:
            raise Exception('Invalid detection state')

    # ...for each image

    return n_negative, n_positive, n_unknown, n_ambiguous

# ..._mark_detection_status()


def is_sas_url(s) -> bool:
    """
    Placeholder for a more robust way to verify that a link is a SAS URL.
    99.999% of the time this will suffice for what we're using it for right now.

    :meta private:
    """

    return (s.startswith(('http://', 'https://')) and ('core.windows.net' in s)
            and ('?' in s))


def relative_sas_url(folder_url, relative_path):
    """
    Given a container-level or folder-level SAS URL, create a SAS URL to the
    specified relative path.

    :meta private:
    """

    relative_path = relative_path.replace('%','%25')
    relative_path = relative_path.replace('#','%23')
    relative_path = relative_path.replace(' ','%20')

    if not is_sas_url(folder_url):
        return None
    tokens = folder_url.split('?')
    assert len(tokens) == 2
    if not tokens[0].endswith('/'):
        tokens[0] = tokens[0] + '/'
    if relative_path.startswith('/'):
        relative_path = relative_path[1:]
    return tokens[0] + relative_path + '?' + tokens[1]


def _render_bounding_boxes(
        image_base_dir,
        image_relative_path,
        display_name,
        detections,
        category_string,
        ground_truth_boxes=None,
        detection_categories=None,
        classification_categories=None,
        options=None):
    """
    Renders detection bounding boxes on a single image.

    This is an internal function; if you want tools for rendering boxes on images, see
    visualization.visualization_utils.

    The source image is:

        image_base_dir / image_relative_path

    The target image is, for example:

        [options.output_dir] /
        ['detections' or 'non_detections'] /
        [filename with slashes turned into tildes]

    "category_string" is a result type, e.g. "detections", "non-detections", "class_coyote"; this
    determines the output folder for the rendered image.

    Only very preliminary support is provided for ground truth box rendering.

    Returns the html info struct for this image in the format that's used for
    write_html_image_list.

    :meta private:
    """

    if options is None:
        options = PostProcessingOptions()

    image_full_path = None

    if category_string in options.rendering_bypass_sets:

        sample_name = category_string + '_' + path_utils.flatten_path(image_relative_path)

    else:

        if is_sas_url(image_base_dir):
            image_full_path = relative_sas_url(image_base_dir, image_relative_path)
        else:
            image_full_path = os.path.join(image_base_dir, image_relative_path)

        # os.path.isfile() is slow when mounting remote directories; much faster
        # to just try/except on the image open.
        try:
            image = vis_utils.open_image(image_full_path)
        except Exception as e:
            print('Warning: could not open image file {}: {}'.format(image_full_path,str(e)))
            image = None
            # return ''

        # Render images to a flat folder
        sample_name = category_string + '_' + path_utils.flatten_path(image_relative_path)
        fullpath = os.path.join(options.output_dir, category_string, sample_name)

        if image is not None:

            original_size = image.size

            # Resize the image if necessary
            if options.viz_target_width is not None:
                image = vis_utils.resize_image(image, options.viz_target_width)

            # Render ground truth boxes if necessary
            if ground_truth_boxes is not None and len(ground_truth_boxes) > 0:

                # Create class labels like "gt_1" or "gt_27"
                gt_classes = [0] * len(ground_truth_boxes)
                label_map = {0:'ground truth'}
                # for i_box,box in enumerate(ground_truth_boxes):
                #    gt_classes.append('_' + str(box[-1]))
                vis_utils.render_db_bounding_boxes(ground_truth_boxes, gt_classes, image,
                                                   original_size=original_size,label_map=label_map,
                                                   thickness=4,expansion=4)

            # Prepare per-category confidence thresholds
            if isinstance(options.confidence_threshold,float):
                rendering_confidence_threshold = options.confidence_threshold
            else:
                category_ids = set()
                for d in detections:
                    category_ids.add(d['category'])
                rendering_confidence_threshold = {}
                for category_id in category_ids:
                    rendering_confidence_threshold[category_id] = \
                        _get_threshold_for_category_id(category_id, options, detection_categories)

            # Render detection boxes
            vis_utils.render_detection_bounding_boxes(
                detections, image,
                label_map=detection_categories,
                classification_label_map=classification_categories,
                confidence_threshold=rendering_confidence_threshold,
                classification_confidence_threshold=options.classification_confidence_threshold,
                thickness=options.line_thickness,
                expansion=options.box_expansion)

            try:
                image.save(fullpath)
            except OSError as e:
                # errno.ENAMETOOLONG doesn't get thrown properly on Windows, so
                # we awkwardly check against a hard-coded limit
                if (e.errno == errno.ENAMETOOLONG) or (len(fullpath) >= 259):
                    extension = os.path.splitext(sample_name)[1]
                    sample_name = category_string + '_' + str(uuid.uuid4()) + extension
                    image.save(os.path.join(options.output_dir, category_string, sample_name))
                else:
                    raise

        # ...if we successfully opened this image

    # Use forward slashes regardless of os
    file_name = '{}/{}'.format(category_string,sample_name)

    info = {
        'filename': file_name,
        'title': display_name,
        'textStyle':\
         'font-family:verdana,arial,calibri;font-size:80%;text-align:left;margin-top:20;margin-bottom:5'
    }

    # Optionally add links back to the original images
    if options.link_images_to_originals and (image_full_path is not None):

        # Handling special characters in links has been pushed down into
        # write_html_image_list
        #
        # link_target = image_full_path.replace('\\','/')
        # link_target  = urllib.parse.quote(link_target)
        link_target = image_full_path
        info['linkTarget'] = link_target

    return info

# ..._render_bounding_boxes(...)


def _prepare_html_subpages(images_html, output_dir, options=None):
    """
    Write out a series of html image lists, e.g. the "detections" or "non-detections"
    pages.

    image_html is a dictionary mapping an html page name (e.g. "detections_animal") to
    a list of image structs friendly to write_html_image_list.

    Returns a dictionary mapping category names to image counts.
    """

    if options is None:
        options = PostProcessingOptions()

    # Count items in each category
    image_counts = {}
    for category_string, image_this_category in images_html.items():
        image_counts[category_string] = len(image_this_category)

    # Optionally sort by filename before writing to html
    if options.html_sort_order == 'filename':
        images_html_sorted = {}
        for category_string, images_this_category in images_html.items():
            sorted_images_this_category = sorted(images_this_category, key=lambda x: x['filename'])
            images_html_sorted[category_string] = sorted_images_this_category
        images_html = images_html_sorted

    # Optionally sort by confidence before writing to html
    elif options.html_sort_order == 'confidence':
        images_html_sorted = {}
        for category_string, images_this_category in images_html.items():

            if not all(['max_conf' in d for d in images_this_category]):
                print(f"Warning: some elements in the {category_string} page don't have confidence " + \
                        "values, can't sort by confidence")
            else:
                sorted_images_this_category = sorted(images_this_category,
                                                     key=lambda x: x['max_conf'], reverse=True)
                images_html_sorted[category_string] = sorted_images_this_category
                images_html = images_html_sorted

    else:
        assert options.html_sort_order == 'random',\
            'Unrecognized sort order {}'.format(options.html_sort_order)
        images_html_sorted = {}
        for category_string, images_this_category in images_html.items():
            sorted_images_this_category = random.sample(images_this_category,len(images_this_category))
            images_html_sorted[category_string] = sorted_images_this_category
        images_html = images_html_sorted

    # Write the individual HTML files
    #
    # category_string will be, e.g., "detections", "non-detections", "class_coyote"
    for category_string, images_this_category in images_html.items():

        html_image_list_options = {}
        html_image_list_options['maxFiguresPerHtmlFile'] = options.max_figures_per_html_file
        html_image_list_options['headerHtml'] = '<h1>{}</h1>'.format(category_string.upper())
        html_image_list_options['pageTitle'] = '{}'.format(category_string.lower())

        # Don't write empty pages
        if len(images_this_category) == 0:
            continue
        else:
            html_filename_base = clean_filename(category_string,replace_whitespace='_')
            write_html_image_list(
                filename=os.path.join(output_dir, '{}.html'.format(html_filename_base)),
                images=images_this_category,
                options=html_image_list_options)

    # ...for each HTML page

    return image_counts

# ..._prepare_html_subpages()


def _get_threshold_for_category_name(category_name,options):
    """
    Determines the confidence threshold we should use for a specific category name.
    """

    if isinstance(options.confidence_threshold,float):
        return options.confidence_threshold
    else:
        assert isinstance(options.confidence_threshold,dict), \
            'confidence_threshold must either be a float or a dict'

    if category_name in options.confidence_threshold:

        return options.confidence_threshold[category_name]

    else:
        assert 'default' in options.confidence_threshold, \
            'category {} not in confidence_threshold dict, and no default supplied'.format(
                category_name)
        return options.confidence_threshold['default']


def _get_threshold_for_category_id(category_id,options,detection_categories):
    """
    Determines the confidence threshold we should use for a specific category ID.

    [detection_categories] is a dict mapping category IDs to names.
    """

    if isinstance(options.confidence_threshold,float):
        return options.confidence_threshold

    assert category_id in detection_categories, \
        'Invalid category ID {}'.format(category_id)

    category_name = detection_categories[category_id]

    return _get_threshold_for_category_name(category_name,options)


def _get_positive_categories(detections,options,detection_categories):
    """
    Gets a sorted list of unique categories (as string IDs) above the threshold for this image

    [detection_categories] is a dict mapping category IDs to names.
    """

    positive_categories = set()
    for d in detections:
        threshold = _get_threshold_for_category_id(d['category'], options, detection_categories)
        if d['conf'] >= threshold:
            positive_categories.add(d['category'])
    return sorted(positive_categories)


def _has_positive_detection(detections,options,detection_categories):
    """
    Determines whether any positive detections are present in the detection list
    [detections].
    """

    found_positive_detection = False
    for d in detections:
        threshold = _get_threshold_for_category_id(d['category'], options, detection_categories)
        if d['conf'] >= threshold:
            found_positive_detection = True
            break
    return found_positive_detection


def _render_image_no_gt(file_info,
                        detection_categories_to_results_name,
                        detection_categories,
                        classification_categories,
                        options):
    r"""
    Renders an image (with no ground truth information)

    Returns a list of rendering structs, where the first item is a category (e.g. "detections_animal"),
    and  the second is a dict of information needed for rendering.  E.g.:

    [['detections_animal',
    {
     'filename': 'detections_animal/detections_animal_blah~01060415.JPG',
     'title': '<b>Result type</b>: detections_animal,
               <b>Image</b>: blah\\01060415.JPG,
               <b>Max conf</b>: 0.897',
      'textStyle': 'font-family:verdana,arial,calibri;font-size:80%;text-align:left;margin-top:20;margin-bottom:5',
      'linkTarget': 'full_path_to_%5C01060415.JPG'
    }]]

    When no classification data is present, this list will always be length-1.  When
    classification data is present, an image may appear in multiple categories.

    Populates the 'max_conf' field of the first element of the list.

    Returns None if there are any errors.
    """

    image_relative_path = file_info['file']

    # Useful debug snippet
    #
    # if 'filename' in image_relative_path:
    #    import pdb; pdb.set_trace()

    max_conf = file_info['max_detection_conf']
    detections = file_info['detections']

    # Determine whether any positive detections are present (using a threshold that
    # may vary by category)
    found_positive_detection = _has_positive_detection(detections,options,detection_categories)

    detection_status = DetectionStatus.DS_UNASSIGNED
    if found_positive_detection:
        detection_status = DetectionStatus.DS_POSITIVE
    else:
        if options.include_almost_detections:
            if max_conf >= options.almost_detection_confidence_threshold:
                detection_status = DetectionStatus.DS_ALMOST
            else:
                detection_status = DetectionStatus.DS_NEGATIVE
        else:
            detection_status = DetectionStatus.DS_NEGATIVE

    if detection_status == DetectionStatus.DS_POSITIVE:
        if options.separate_detections_by_category:
            positive_categories = tuple(_get_positive_categories(detections,options,detection_categories))
            if positive_categories not in detection_categories_to_results_name:
                raise ValueError('Error: {} not in category mapping (file {})'.format(
                    str(positive_categories),image_relative_path))
            category_string = detection_categories_to_results_name[positive_categories]
        else:
            category_string = 'detections'

    elif detection_status == DetectionStatus.DS_NEGATIVE:
        category_string = 'non_detections'
    else:
        assert detection_status == DetectionStatus.DS_ALMOST
        category_string = 'almost_detections'

    display_name = '<b>Result type</b>: {}, <b>image</b>: {}, <b>max conf</b>: {:0.3f}'.format(
        category_string, image_relative_path, max_conf)

    # Are there any bonus fields we need to include in each image header?
    if options.additional_image_fields_to_display is not None:

        for field_name in options.additional_image_fields_to_display:

            if field_name in file_info:

                field_value = file_info[field_name]

                if (field_value is None) or \
                    (isinstance(field_value,float) and np.isnan(field_value)):
                        continue

                # Optionally use a display name that's different from the field name
                if isinstance(options.additional_image_fields_to_display,dict):
                    field_display_name = \
                        options.additional_image_fields_to_display[field_name]
                else:
                    field_display_name = field_name
                field_string = '<b>{}</b>: {}'.format(field_display_name,field_value)
                display_name += ', {}'.format(field_string)

    rendering_options = copy.copy(options)
    if detection_status == DetectionStatus.DS_ALMOST:
        rendering_options.confidence_threshold = \
            rendering_options.almost_detection_confidence_threshold

    rendered_image_html_info = _render_bounding_boxes(
        image_base_dir=options.image_base_dir,
        image_relative_path=image_relative_path,
        display_name=display_name,
        detections=detections,
        category_string=category_string,
        ground_truth_boxes=None,
        detection_categories=detection_categories,
        classification_categories=classification_categories,
        options=rendering_options)

    image_result = None

    if len(rendered_image_html_info) > 0:

        image_result = [[category_string, rendered_image_html_info]]
        classes_rendered_this_image = set()
        max_conf = 0

        for det in detections:

            if det['conf'] > max_conf:
                max_conf = det['conf']

            # We make the decision here that only "detections" (not "almost-detections")
            # will appear on the classification category pages
            detection_threshold = \
                _get_threshold_for_category_id(det['category'], options, detection_categories)
            if det['conf'] < detection_threshold:
                continue

            if ('classifications' in det) and (len(det['classifications']) > 0) and \
                (category_string != 'non_detections'):

                # This is a list of [class,confidence] pairs, sorted by classification confidence
                classifications = det['classifications']
                top1_class_id = classifications[0][0]
                top1_class_name = classification_categories[top1_class_id]
                top1_class_score = classifications[0][1]

                # If we either don't have a classification confidence threshold, or
                # we've met our classification confidence threshold
                if (options.classification_confidence_threshold < 0) or \
                    (top1_class_score >= options.classification_confidence_threshold):
                    class_string = 'class_{}'.format(top1_class_name)
                else:
                    class_string = 'class_unreliable'

                if class_string not in classes_rendered_this_image:
                    image_result.append([class_string,
                                         rendered_image_html_info])
                    classes_rendered_this_image.add(class_string)

            # ...if this detection has classification info

        # ...for each detection

        image_result[0][1]['max_conf'] = max_conf

    # ...if we got valid rendering info back from _render_bounding_boxes()

    return image_result

# ...def _render_image_no_gt()


def _render_image_with_gt(file_info,
                          ground_truth_indexed_db,
                          detection_categories,
                          classification_categories,
                          options):
    """
    Render an image with ground truth information.  See _render_image_no_gt for return
    data format.
    """

    image_relative_path = file_info['file']
    max_conf = file_info['max_detection_conf']
    detections = file_info['detections']

    # This should already have been normalized to either '/' or '\'

    image_id = ground_truth_indexed_db.filename_to_id.get(image_relative_path, None)
    if image_id is None:
        print('Warning: couldn''t find ground truth for image {}'.format(image_relative_path))
        return None

    image = ground_truth_indexed_db.image_id_to_image[image_id]
    annotations = ground_truth_indexed_db.image_id_to_annotations[image_id]

    ground_truth_boxes = []
    for ann in annotations:
        if 'bbox' in ann:
            ground_truth_box = [x for x in ann['bbox']]
            ground_truth_box.append(ann['category_id'])
            ground_truth_boxes.append(ground_truth_box)

    gt_status = image['_detection_status']

    gt_presence = bool(gt_status)

    gt_classes = CameraTrapJsonUtils.annotations_to_class_names(
        annotations, ground_truth_indexed_db.cat_id_to_name)
    gt_class_summary = ','.join(gt_classes)

    if gt_status > DetectionStatus.DS_MAX_DEFINITIVE_VALUE:
        print(f'Skipping image {image_id}, does not have a definitive '
              f'ground truth status (status: {gt_status}, classes: {gt_class_summary})')
        return None

    detected = _has_positive_detection(detections, options, detection_categories)

    if gt_presence and detected:
        if '_classification_accuracy' not in image.keys():
            result_type = 'tp'
        elif np.isclose(1, image['_classification_accuracy']):
            result_type = 'tpc'
        else:
            result_type = 'tpi'
    elif not gt_presence and detected:
        result_type = 'fp'
    elif gt_presence and not detected:
        result_type = 'fn'
    else:
        result_type = 'tn'

    display_name = '<b>Result type</b>: {}, <b>Presence</b>: {}, <b>Class</b>: {}, <b>Max conf</b>: {:0.3f}%, <b>Image</b>: {}'.format( # noqa
        result_type.upper(),
        str(gt_presence),
        gt_class_summary,
        max_conf * 100,
        image_relative_path)

    rendered_image_html_info = _render_bounding_boxes(
        image_base_dir=options.image_base_dir,
        image_relative_path=image_relative_path,
        display_name=display_name,
        detections=detections,
        category_string=result_type,
        ground_truth_boxes=ground_truth_boxes,
        detection_categories=detection_categories,
        classification_categories=classification_categories,
        options=options)

    image_result = None
    if len(rendered_image_html_info) > 0:
        image_result = [[result_type, rendered_image_html_info]]
        for gt_class in gt_classes:
            image_result.append(['class_{}'.format(gt_class), rendered_image_html_info])

    return image_result

# ...def _render_image_with_gt()


#%% Main function

def process_batch_results(options):
    """
    Given a .json or .csv file containing MD results, do one or more of the following:

    * Sample detections/non-detections and render to HTML (when ground truth isn't
      available) (this is 99.9% of what this module is for)
    * Evaluate detector precision/recall, optionally rendering results (requires
      ground truth)
    * Sample true/false positives/negatives and render to HTML (requires ground
      truth)

    Ground truth, if available, must be in COCO Camera Traps format:

    https://github.com/agentmorris/MegaDetector/blob/main/megadetector/data_management/README.md#coco-camera-traps-format

    Args:
        options (PostProcessingOptions): everything we need to render a preview/analysis for
            this set of results; see the PostProcessingOptions class for details.

    Returns:
        PostProcessingResults: information about the results/preview, most importantly the
        HTML filename of the output.  See the PostProcessingResults class for details.
    """
    ppresults = PostProcessingResults()

    ##%% Expand some options for convenience

    output_dir = options.output_dir


    ##%% Prepare output dir

    os.makedirs(output_dir, exist_ok=True)


    ##%% Load ground truth if available

    ground_truth_indexed_db = None

    if (options.ground_truth_json_file is not None) and (len(options.ground_truth_json_file) > 0):
        assert (options.confidence_threshold is None) or (isinstance(options.confidence_threshold,float)), \
            'Variable confidence thresholds are not supported when supplying ground truth'

    if (options.ground_truth_json_file is not None) and (len(options.ground_truth_json_file) > 0):

        if options.separate_detections_by_category:
            print("Warning: I don't know how to separate categories yet when doing " + \
                  "a P/R analysis, disabling category separation")
            options.separate_detections_by_category = False

        ground_truth_indexed_db = IndexedJsonDb(
            options.ground_truth_json_file,
            b_normalize_paths=True,
            filename_replacements=options.ground_truth_filename_replacements)

        # Mark images in the ground truth as positive or negative
        n_negative, n_positive, n_unknown, n_ambiguous = _mark_detection_status(
            ground_truth_indexed_db,
            negative_classes=options.negative_classes,
            unknown_classes=options.unlabeled_classes)

        print(f'Finished loading and indexing ground truth: {n_negative} '
              f'negative, {n_positive} positive, {n_unknown} unknown, '
              f'{n_ambiguous} ambiguous')

        if n_positive == 0:
            print('\n*** Warning: no positives found in ground truth, analysis won\'t be very meaningful ***\n')
        if n_negative == 0:
            print('\n*** Warning: no negatives found in ground truth, analysis won\'t be very meaningful ***\n')
        if n_ambiguous > 0:
            print('\n*** Warning: {} images with ambiguous positive/negative status found in ground truth ***\n'.format(
                n_ambiguous))


    ##%% Load detection (and possibly classification) results

    # If the caller hasn't supplied results, load them
    if options.api_detection_results is None:
        detections_df, other_fields = load_api_results(
            options.md_results_file, force_forward_slashes=True,
            filename_replacements=options.api_output_filename_replacements)
        ppresults.api_detection_results = detections_df
        ppresults.api_other_fields = other_fields

    else:
        print('Bypassing detection results loading...')
        assert options.api_other_fields is not None
        detections_df = options.api_detection_results
        other_fields = options.api_other_fields

    # Determine confidence thresholds if necessary

    if options.confidence_threshold is None:
        options.confidence_threshold = \
            get_typical_confidence_threshold_from_results(other_fields)
        print('Choosing default confidence threshold of {} based on MD version'.format(
            options.confidence_threshold))

    if options.almost_detection_confidence_threshold is None and options.include_almost_detections:
        assert isinstance(options.confidence_threshold,float), \
            'If you are using a dictionary of confidence thresholds and almost-detections are enabled, ' + \
            'you need to supply a threshold for almost detections.'
        options.almost_detection_confidence_threshold = options.confidence_threshold - 0.05
        if options.almost_detection_confidence_threshold < 0:
            options.almost_detection_confidence_threshold = 0

    # Remove rows with inference failures (typically due to corrupt images)
    n_failures = 0
    if 'failure' in detections_df.columns:
        n_failures = detections_df['failure'].count()
        print('Ignoring {} failed images'.format(n_failures))
        # Explicitly forcing a copy() operation here to suppress "trying to be set
        # on a copy" warnings (and associated risks) below.
        detections_df = detections_df[detections_df['failure'].isna()].copy()

    assert other_fields is not None

    detection_categories = other_fields['detection_categories']

    # Convert keys and values to lowercase
    classification_categories = other_fields.get('classification_categories', {})
    if classification_categories is not None:
        classification_categories = {
            k.lower(): v.lower()
            for k, v in classification_categories.items()
        }

    # Count detections and almost-detections for reporting purposes
    n_positives = 0
    n_almosts = 0

    print('Assigning images to rendering categories')

    for i_row,row in tqdm(detections_df.iterrows(),total=len(detections_df)):

        detections = row['detections']
        max_conf = row['max_detection_conf']
        if _has_positive_detection(detections, options, detection_categories):
            n_positives += 1
        elif (options.almost_detection_confidence_threshold is not None) and \
             (max_conf >= options.almost_detection_confidence_threshold):
            n_almosts += 1

    print(f'Finished loading and preprocessing {len(detections_df)} rows '
          f'from detector output, predicted {n_positives} positives.')

    if options.include_almost_detections:
        print('...and {} almost-positives'.format(n_almosts))


    ##%% Find descriptive metadata to include at the top of the page

    if options.job_name_string is not None:
        job_name_string = options.job_name_string
    else:
        # This is rare; it only happens during debugging when the caller
        # is supplying already-loaded MD results.
        if options.md_results_file is None:
            job_name_string = 'unknown'
        else:
            job_name_string = os.path.basename(options.md_results_file)

    if options.model_version_string is not None:
        model_version_string = options.model_version_string
    else:

        if 'info' not in other_fields or 'detector' not in other_fields['info']:
            print('No model metadata supplied, assuming MDv4')
            model_version_string = 'MDv4 (assumed)'
        else:
            model_version_string = other_fields['info']['detector']


    ##%% If we have ground truth, remove images we can't match to ground truth

    if ground_truth_indexed_db is not None:

        b_match = detections_df['file'].isin(
            ground_truth_indexed_db.filename_to_id)
        print(f'Confirmed filename matches to ground truth for {sum(b_match)} '
              f'of {len(detections_df)} files')

        detections_df = detections_df[b_match]
        detector_files = detections_df['file'].tolist()

        assert len(detector_files) > 0, (
            'No detection files available, possible path issue?')

        print('Trimmed detection results to {} files'.format(len(detector_files)))


    ##%% (Optionally) sample from the full set of images

    images_to_visualize = detections_df

    if (options.num_images_to_sample is not None) and (options.num_images_to_sample > 0):
        images_to_visualize = images_to_visualize.sample(
            n=min(options.num_images_to_sample, len(images_to_visualize)),
            random_state=options.sample_seed)

    output_html_file = ''

    style_header = """<head>
        <title>Detection results preview</title>
        <style type="text/css">
        a { text-decoration: none; }
        body { font-family: segoe ui, calibri, "trebuchet ms", verdana, arial, sans-serif; }
        div.contentdiv { margin-left: 20px; }
        </style>
        </head>"""


    ##%% Fork here depending on whether or not ground truth is available

    # If we have ground truth, we'll compute precision/recall and sample tp/fp/tn/fn.
    #
    # Otherwise we'll just visualize detections/non-detections.

    if ground_truth_indexed_db is not None:

        ##%% Detection evaluation: compute precision/recall

        # numpy array of maximum confidence values
        p_detection = detections_df['max_detection_conf'].values
        n_detection_values = len(p_detection)

        # numpy array of bools (0.0/1.0), and -1 as null value
        gt_detections = np.zeros(n_detection_values, dtype=float)

        n_positive = 0
        n_negative = 0

        for i_detection, fn in enumerate(detector_files):

            image_id = ground_truth_indexed_db.filename_to_id[fn]
            image = ground_truth_indexed_db.image_id_to_image[image_id]
            detection_status = image['_detection_status']

            if detection_status == DetectionStatus.DS_NEGATIVE:
                gt_detections[i_detection] = 0.0
                n_negative += 1
            elif detection_status == DetectionStatus.DS_POSITIVE:
                gt_detections[i_detection] = 1.0
                n_positive += 1
            else:
                gt_detections[i_detection] = -1.0

        print('Of {} ground truth values, found {} positives and {} negatives'.format(
            len(detections_df),n_positive,n_negative))

        # Don't include ambiguous/unknown ground truth in precision/recall analysis
        b_valid_ground_truth = gt_detections >= 0.0

        p_detection_pr = p_detection[b_valid_ground_truth]
        gt_detections_pr = (gt_detections[b_valid_ground_truth] == 1.)

        print('Including {} of {} values in p/r analysis'.format(np.sum(b_valid_ground_truth),
              len(b_valid_ground_truth)))

        precisions, recalls, thresholds = precision_recall_curve(gt_detections_pr, p_detection_pr)

        # For completeness, include the result at a confidence threshold of 1.0
        thresholds = np.append(thresholds, [1.0])

        precisions_recalls = pd.DataFrame(data={
            'confidence_threshold': thresholds,
            'precision': precisions,
            'recall': recalls
        })

        # Compute and print summary statistics
        average_precision = average_precision_score(gt_detections_pr, p_detection_pr)
        print('Average precision: {:.1%}'.format(average_precision))

        # Thresholds go up throughout precisions/recalls/thresholds; find the last
        # value where recall is at or above target.  That's our precision @ target recall.

        i_above_target_recall = (np.where(recalls >= options.target_recall))

        # np.where returns a tuple of arrays, but in this syntax where we're
        # comparing an array with a scalar, there will only be one element.
        assert len (i_above_target_recall) == 1

        # Convert back to a list
        i_above_target_recall = i_above_target_recall[0].tolist()

        if len(i_above_target_recall) == 0:
            precision_at_target_recall = 0.0
        else:
            precision_at_target_recall = precisions[i_above_target_recall[-1]]
        print('Precision at {:.1%} recall: {:.1%}'.format(options.target_recall,
                                                          precision_at_target_recall))

        cm_predictions = np.array(p_detection_pr) > options.confidence_threshold
        cm = confusion_matrix(gt_detections_pr, cm_predictions, labels=[False,True])

        # Flatten the confusion matrix
        tn, fp, fn, tp = cm.ravel()

        precision_at_confidence_threshold = tp / (tp + fp)
        recall_at_confidence_threshold = tp / (tp + fn)
        f1 = 2.0 * (precision_at_confidence_threshold * recall_at_confidence_threshold) / \
            (precision_at_confidence_threshold + recall_at_confidence_threshold)

        print('At a confidence threshold of {:.1%}, precision={:.1%}, recall={:.1%}, f1={:.1%}'.format(
                options.confidence_threshold, precision_at_confidence_threshold,
                recall_at_confidence_threshold, f1))

        ##%% Collect classification results, if they exist

        classifier_accuracies = []

        # Mapping of classnames to idx for the confusion matrix.
        #
        # The lambda is actually kind of a hack, because we use assume that
        # the following code does not reassign classname_to_idx
        classname_to_idx = collections.defaultdict(lambda: len(classname_to_idx))

        # Confusion matrix as defaultdict of defaultdict
        #
        # Rows / first index is ground truth, columns / second index is predicted category
        classifier_cm = collections.defaultdict(lambda: collections.defaultdict(lambda: 0))

        # i_detection = 0; fn = detector_files[i_detection]; print(fn)
        assert len(detector_files) == len(detections_df)
        for i_detection, fn in enumerate(detector_files):

            image_id = ground_truth_indexed_db.filename_to_id[fn]
            image = ground_truth_indexed_db.image_id_to_image[image_id]
            detections = detections_df['detections'].iloc[i_detection]
            pred_class_ids = [det['classifications'][0][0] \
                for det in detections if 'classifications' in det.keys()]
            pred_classnames = [classification_categories[pd] for pd in pred_class_ids]

            # If this image has classification predictions, and an unambiguous class
            # annotated, and is a positive image...
            if len(pred_classnames) > 0 \
                    and '_unambiguous_category' in image.keys() \
                    and image['_detection_status'] == DetectionStatus.DS_POSITIVE:

                # The unambiguous category, we make this a set for easier handling afterward
                gt_categories = set([image['_unambiguous_category']])
                pred_categories = set(pred_classnames)

                # Compute the accuracy as intersection of union,
                # i.e. (# of categories in both prediction and GT)
                #      divided by (# of categories in either prediction or GT
                #
                # In case of only one GT category, the result will be 1.0, if
                # prediction is one category and this category matches GT
                #
                # It is 1.0/(# of predicted top-1 categories), if the GT is
                # one of the predicted top-1 categories.
                #
                # It is 0.0, if none of the predicted categories is correct

                classifier_accuracies.append(
                    len(gt_categories & pred_categories)
                    / len(gt_categories | pred_categories)
                )
                image['_classification_accuracy'] = classifier_accuracies[-1]

                # Distribute this accuracy across all predicted categories in the
                # confusion matrix
                assert len(gt_categories) == 1
                gt_class_idx = classname_to_idx[list(gt_categories)[0]]
                for pred_category in pred_categories:
                    pred_class_idx = classname_to_idx[pred_category]
                    classifier_cm[gt_class_idx][pred_class_idx] += 1

        # ...for each file in the detection results

        # If we have classification results
        if len(classifier_accuracies) > 0:

            # Build confusion matrix as array from classifier_cm
            all_class_ids = sorted(classname_to_idx.values())
            classifier_cm_array = np.array(
                [[classifier_cm[r_idx][c_idx] for c_idx in all_class_ids] for \
                 r_idx in all_class_ids], dtype=float)
            classifier_cm_array /= (classifier_cm_array.sum(axis=1, keepdims=True) + 1e-7)

            # Print some statistics
            print('Finished computation of {} classification results'.format(
                len(classifier_accuracies)))
            print('Mean accuracy: {}'.format(np.mean(classifier_accuracies)))

            # Prepare confusion matrix output

            # Get confusion matrix as string
            sio = io.StringIO()
            np.savetxt(sio, classifier_cm_array * 100, fmt='%5.1f')
            cm_str = sio.getvalue()
            # Get fixed-size classname for each idx
            idx_to_classname = {v:k for k,v in classname_to_idx.items()}
            classname_list = [idx_to_classname[idx] for idx in sorted(classname_to_idx.values())]
            classname_headers = ['{:<5}'.format(cname[:5]) for cname in classname_list]

            # Prepend class name on each line and add to the top
            cm_str_lines = [' ' * 16 + ' '.join(classname_headers)]
            cm_str_lines += ['{:>15}'.format(cn[:15]) + ' ' + cm_line for cn, cm_line in \
                             zip(classname_list, cm_str.splitlines(), strict=True)]

            # Print formatted confusion matrix
            if False:
                # Actually don't, this gets really messy in all but the widest consoles
                print('Confusion matrix: ')
                print(*cm_str_lines, sep='\n')

            # Plot confusion matrix

            # To manually add more space at bottom: plt.rcParams['figure.subplot.bottom'] = 0.1
            #
            # Add 0.5 to figsize for every class. For two classes, this will result in
            # fig = plt.figure(figsize=[4,4])
            fig = plot_utils.plot_confusion_matrix(
                classifier_cm_array,
                classname_list,
                normalize=False,
                title='Confusion matrix',
                cmap=plt.cm.Blues,
                vmax=1.0,
                use_colorbar=True,
                y_label=True)
            cm_figure_relative_filename = 'confusion_matrix.png'
            cm_figure_filename = os.path.join(output_dir, cm_figure_relative_filename)
            plt.savefig(cm_figure_filename)
            plt.close(fig)

        # ...if we have classification results


        ##%% Render output

        # Write p/r table to .csv file in output directory
        pr_table_filename = os.path.join(output_dir, 'prec_recall.csv')
        precisions_recalls.to_csv(pr_table_filename, index=False)

        # Write precision/recall plot to .png file in output directory
        t = 'Precision-Recall curve: AP={:0.1%}, P@{:0.1%}={:0.1%}'.format(
            average_precision, options.target_recall, precision_at_target_recall)
        fig = plot_utils.plot_precision_recall_curve(precisions, recalls, t)

        pr_figure_relative_filename = 'prec_recall.png'
        pr_figure_filename = os.path.join(output_dir, pr_figure_relative_filename)
        fig.savefig(pr_figure_filename)
        plt.close(fig)


        ##%% Sampling

        # Sample true/false positives/negatives with correct/incorrect top-1
        # classification and render to html

        # Accumulate html image structs (in the format expected by write_html_image_list)
        # for each category, e.g. 'tp', 'fp', ..., 'class_bird', ...
        images_html = collections.defaultdict(list)

        # Add default entries by accessing them for the first time
        [images_html[res] for res in ['tp', 'tpc', 'tpi', 'fp', 'tn', 'fn']]
        for res in images_html.keys():
            os.makedirs(os.path.join(output_dir, res), exist_ok=True)

        image_count = len(images_to_visualize)

        # Each element will be a list of 2-tuples, with elements [collection name,html info struct]
        rendering_results = []

        # Each element will be a three-tuple with elements file,max_conf,detections
        files_to_render = []

        # Assemble the information we need for rendering, so we can parallelize without
        # dealing with Pandas
        # i_row = 0; row = images_to_visualize.iloc[0]
        for _, row in images_to_visualize.iterrows():

            # Filenames should already have been normalized to either '/' or '\'
            files_to_render.append(row.to_dict())

        start_time = time.time()
        if options.parallelize_rendering:
            pool = None
            try:
                if options.parallelize_rendering_n_cores is None:
                    if options.parallelize_rendering_with_threads:
                        pool = ThreadPool()
                    else:
                        pool = Pool()
                else:
                    if options.parallelize_rendering_with_threads:
                        pool = ThreadPool(options.parallelize_rendering_n_cores)
                        worker_string = 'threads'
                    else:
                        pool = Pool(options.parallelize_rendering_n_cores)
                        worker_string = 'processes'
                    print('Rendering images with {} {}'.format(options.parallelize_rendering_n_cores,
                                                               worker_string))

                rendering_results = list(tqdm(pool.imap(
                    partial(_render_image_with_gt,
                            ground_truth_indexed_db=ground_truth_indexed_db,
                            detection_categories=detection_categories,
                            classification_categories=classification_categories,
                            options=options),
                    files_to_render), total=len(files_to_render)))
            finally:
                if pool is not None:
                    pool.close()
                    pool.join()
                    print('Pool closed and joined for GT rendering')
        else:
            for file_info in tqdm(files_to_render):
                rendering_results.append(_render_image_with_gt(
                    file_info,ground_truth_indexed_db,
                    detection_categories,classification_categories,
                    options=options))
        elapsed = time.time() - start_time

        # Map all the rendering results in the list rendering_results into the
        # dictionary images_html, which maps category names to lists of results
        image_rendered_count = 0
        for rendering_result in rendering_results:
            if rendering_result is None:
                continue
            image_rendered_count += 1
            for assignment in rendering_result:
                images_html[assignment[0]].append(assignment[1])

        # Prepare the individual html image files
        image_counts = _prepare_html_subpages(images_html, output_dir, options)

        print('{} images rendered (of {})'.format(image_rendered_count,image_count))

        # Write index.html
        all_tp_count = image_counts['tp'] + image_counts['tpc'] + image_counts['tpi']
        total_count = all_tp_count + image_counts['tn'] + image_counts['fp'] + image_counts['fn']

        classification_detection_results = """&nbsp;&nbsp;&nbsp;&nbsp;<a href="tpc.html">with all correct top-1 predictions (TPC)</a> ({})<br/>
           &nbsp;&nbsp;&nbsp;&nbsp;<a href="tpi.html">with one or more incorrect top-1 prediction (TPI)</a> ({})<br/>
           &nbsp;&nbsp;&nbsp;&nbsp;<a href="tp.html">without classification evaluation</a><sup>*</sup> ({})<br/>""".format(
            image_counts['tpc'],
            image_counts['tpi'],
            image_counts['tp']
        )

        confidence_threshold_string = ''
        if isinstance(options.confidence_threshold,float):
            confidence_threshold_string = '{:.2%}'.format(options.confidence_threshold)
        else:
            confidence_threshold_string = str(options.confidence_threshold)

        index_page = """<html>
        {}
        <body>
        <h2>Evaluation</h2>

        <h3>Job metadata</h3>

        <div class="contentdiv">
        <p>Job name: {}<br/>
        <p>Model version: {}</p>
        </div>

        <h3>Sample images</h3>
        <div class="contentdiv">
        <p>A sample of {} images, annotated with detections above confidence {}.</p>
        <a href="tp.html">True positives (TP)</a> ({}) ({:0.1%})<br/>
        CLASSIFICATION_PLACEHOLDER_1
        <a href="tn.html">True negatives (TN)</a> ({}) ({:0.1%})<br/>
        <a href="fp.html">False positives (FP)</a> ({}) ({:0.1%})<br/>
        <a href="fn.html">False negatives (FN)</a> ({}) ({:0.1%})<br/>
        CLASSIFICATION_PLACEHOLDER_2
        </div>
        """.format(
            style_header,job_name_string,model_version_string,
            image_count, confidence_threshold_string,
            all_tp_count, all_tp_count/total_count,
            image_counts['tn'], image_counts['tn']/total_count,
            image_counts['fp'], image_counts['fp']/total_count,
            image_counts['fn'], image_counts['fn']/total_count
        )

        index_page += """
            <h3>Detection results</h3>
            <div class="contentdiv">
            <p>At a confidence threshold of {}, precision={:0.1%}, recall={:0.1%}</p>
            <p><strong>Precision/recall summary for all {} images</strong></p><img src="{}"><br/>
            </div>
            """.format(
                confidence_threshold_string, precision_at_confidence_threshold, recall_at_confidence_threshold,
                len(detections_df), pr_figure_relative_filename
           )

        if len(classifier_accuracies) > 0:
            index_page = index_page.replace('CLASSIFICATION_PLACEHOLDER_1',classification_detection_results)
            index_page = index_page.replace('CLASSIFICATION_PLACEHOLDER_2',"""<p><sup>*</sup>We do not evaluate the classification result of images
                if the classification information is missing, if the image contains
                categories like &lsquo;empty&rsquo; or &lsquo;human&rsquo;, or if the image has multiple
                classification labels.</p>""")
        else:
            index_page = index_page.replace('CLASSIFICATION_PLACEHOLDER_1','')
            index_page = index_page.replace('CLASSIFICATION_PLACEHOLDER_2','')

        if len(classifier_accuracies) > 0:
            index_page += """
                <h3>Classification results</h3>
                <div class="contentdiv">
                <p>Classification accuracy: {:.2%}<br>
                The accuracy is computed only for images with exactly one classification label.
                The accuracy of an image is computed as 1/(number of unique detected top-1 classes),
                i.e. if the model detects multiple boxes with different top-1 classes, then the accuracy
                decreases and the image is put into 'TPI'.</p>
                <p>Confusion matrix:</p>
                <p><img src="{}"></p>
                <div style='font-family:monospace;display:block;'>{}</div>
                </div>
                """.format(
                    np.mean(classifier_accuracies),
                    cm_figure_relative_filename,
                    "<br>".join(cm_str_lines).replace(' ', '&nbsp;')
                )

        # Show links to each GT class
        #
        # We could do this without classification results; currently we don't.
        if len(classname_to_idx) > 0:

            index_page += '<h3>Images of specific classes</h3><br/><div class="contentdiv">'
            # Add links to all available classes
            for cname in sorted(classname_to_idx.keys()):
                cname_print_string = cname
                if len(cname_print_string) == 0:
                    cname_print_string = '[no name available]'
                index_page += '<a href="class_{0}.html">{0}</a> ({1})<br>'.format(
                    cname_print_string,
                    len(images_html['class_{}'.format(cname)]))
            index_page += '</div>'

        # Write custom footer if it was provided
        if (options.footer_text is not None) and (len(options.footer_text) > 0):
            index_page += '{}\n'.format(options.footer_text)

        # Close open html tags
        index_page += '\n</body></html>\n'

        output_html_file = os.path.join(output_dir, 'index.html')
        with open(output_html_file, 'w',
                  encoding=options.output_html_encoding) as f:
            f.write(index_page)

        print('Finished writing html to {}'.format(output_html_file))

    # ...if we have ground truth


    ##%% Otherwise, if we don't have ground truth...

    else:

        ##%% Sample detections/non-detections

        # Accumulate html image structs (in the format expected by write_html_image_list)
        # for each category
        images_html = collections.defaultdict(list)

        # Add default entries by accessing them for the first time

        # Maps sorted tuples of detection category IDs (string ints) - e.g. ("1"), ("1", "4", "7") - to
        # result set names, e.g. "detections_human", "detections_cat_truck".
        detection_categories_to_results_name = {}

        # Keep track of which categories are single-class (e.g. "animal") and which are
        # combinations (e.g. "animal_vehicle")
        detection_categories_to_category_count = {}

        # For the creation of a "non-detections" category
        images_html['non_detections']
        detection_categories_to_category_count['non_detections'] = 0


        if not options.separate_detections_by_category:
            # For the creation of a "detections" category
            images_html['detections']
            detection_categories_to_category_count['detections'] = 0
        else:
            # Add a set of results for each category and combination of categories, e.g.
            # "detections_animal_vehicle".  When we're using this script for non-MegaDetector
            # results, this can generate lots of categories, e.g. detections_bear_bird_cat_dog_pig.
            # We'll keep that huge set of combinations in this map, but we'll only write
            # out links for the ones that are non-empty.
            used_combinations = set()

            # row = images_to_visualize.iloc[0]
            for i_row, row in images_to_visualize.iterrows():
                detections_this_row = row['detections']
                above_threshold_category_ids_this_row = set()
                for detection in detections_this_row:
                    threshold = _get_threshold_for_category_id(detection['category'], options, detection_categories)
                    if detection['conf'] >= threshold:
                        above_threshold_category_ids_this_row.add(detection['category'])
                if len(above_threshold_category_ids_this_row) == 0:
                    continue
                sorted_categories_this_row = tuple(sorted(above_threshold_category_ids_this_row))
                used_combinations.add(sorted_categories_this_row)

            for sorted_subset in used_combinations:
                assert len(sorted_subset) > 0
                results_name = 'detections'
                for category_id in sorted_subset:
                    results_name = results_name + '_' + detection_categories[category_id]
                images_html[results_name]
                detection_categories_to_results_name[sorted_subset] = results_name
                detection_categories_to_category_count[results_name] = len(sorted_subset)

        if options.include_almost_detections:
            images_html['almost_detections']
            detection_categories_to_category_count['almost_detections'] = 0

        # Create output directories
        for res in images_html.keys():
            os.makedirs(os.path.join(output_dir, res), exist_ok=True)

        image_count = len(images_to_visualize)

        # Each element will be a list of 2-tuples, with elements [collection name,html info struct]
        rendering_results = []

        # list of 3-tuples with elements (file, max_conf, detections)
        files_to_render = []

        # Assemble the information we need for rendering, so we can parallelize without
        # dealing with Pandas
        # i_row = 0; row = images_to_visualize.iloc[0]
        for _, row in images_to_visualize.iterrows():

            assert isinstance(row['detections'],list)

            # Filenames should already have been normalized to either '/' or '\'
            files_to_render.append(row.to_dict())

        start_time = time.time()
        if options.parallelize_rendering:
            pool = None
            try:
                if options.parallelize_rendering_n_cores is None:
                    if options.parallelize_rendering_with_threads:
                        pool = ThreadPool()
                    else:
                        pool = Pool()
                else:
                    if options.parallelize_rendering_with_threads:
                        pool = ThreadPool(options.parallelize_rendering_n_cores)
                        worker_string = 'threads'
                    else:
                        pool = Pool(options.parallelize_rendering_n_cores)
                        worker_string = 'processes'
                    print('Rendering images with {} {}'.format(options.parallelize_rendering_n_cores,
                                                               worker_string))

                # _render_image_no_gt(file_info,detection_categories_to_results_name,
                # detection_categories,classification_categories)

                rendering_results = list(tqdm(pool.imap(
                    partial(_render_image_no_gt,
                            detection_categories_to_results_name=detection_categories_to_results_name,
                            detection_categories=detection_categories,
                            classification_categories=classification_categories,
                            options=options),
                            files_to_render), total=len(files_to_render)))
            finally:
                if pool is not None:
                    pool.close()
                    pool.join()
                    print('Pool closed and joined for non-GT rendering')
        else:
            for file_info in tqdm(files_to_render):
                rendering_result = _render_image_no_gt(file_info,
                                                       detection_categories_to_results_name,
                                                       detection_categories,
                                                       classification_categories,
                                                       options=options)
                rendering_results.append(rendering_result)

        elapsed = time.time() - start_time

        # Do we have classification results in addition to detection results?
        has_classification_info = False

        # Map all the rendering results in the list rendering_results into the
        # dictionary images_html
        image_rendered_count = 0
        for rendering_result in rendering_results:
            if rendering_result is None:
                continue
            image_rendered_count += 1
            for assignment in rendering_result:
                if 'class' in assignment[0]:
                    has_classification_info = True
                images_html[assignment[0]].append(assignment[1])

        # Prepare the individual html image files
        image_counts = _prepare_html_subpages(images_html, output_dir, options)

        if image_rendered_count == 0:
            seconds_per_image = 0.0
        else:
            seconds_per_image = elapsed/image_rendered_count

        print('Rendered {} images (of {}) in {} ({} per image)'.format(image_rendered_count,
              image_count,humanfriendly.format_timespan(elapsed),
              humanfriendly.format_timespan(seconds_per_image)))

        # Write index.html

        # We can't just sum these, because image_counts includes images in both their
        # detection and classification classes
        total_images = 0
        for k in image_counts.keys():
            v = image_counts[k]
            if has_classification_info and k.startswith('class_'):
                continue
            total_images += v

        if total_images != image_count:
            print('Warning, missing images: image_count is {}, total_images is {}'.format(total_images,image_count))

        almost_detection_string = ''
        if options.include_almost_detections:
            almost_detection_string = ' (&ldquo;almost detection&rdquo; threshold at {:.1%})'.format(
                options.almost_detection_confidence_threshold)

        confidence_threshold_string = ''
        if isinstance(options.confidence_threshold,float):
            confidence_threshold_string = '{:.2%}'.format(options.confidence_threshold)
        else:
            confidence_threshold_string = str(options.confidence_threshold)

        index_page = """<html>\n{}\n<body>\n
        <h2>Visualization of results for {}</h2>\n
        <p>A sample of {} images (of {} total)FAILURE_PLACEHOLDER, annotated with detections above confidence {}{}.</p>\n

        <div class="contentdiv">
        <p>Model version: {}</p>
        </div>

        <h3>Detection results</h3>\n
        <div class="contentdiv">\n""".format(
            style_header, job_name_string, image_count, len(detections_df), confidence_threshold_string,
            almost_detection_string, model_version_string)

        failure_string = ''
        if n_failures is not None:
            failure_string = ' ({} failures)'.format(n_failures)
        index_page = index_page.replace('FAILURE_PLACEHOLDER',failure_string)

        def result_set_name_to_friendly_name(result_set_name):
            friendly_name = ''
            friendly_name = result_set_name.replace('_','-')
            if friendly_name.startswith('detections-'):
                friendly_name = friendly_name.replace('detections-', 'detections: ')
            friendly_name = friendly_name.capitalize()
            return friendly_name

        sorted_result_set_names = sorted(list(images_html.keys()))

        result_set_name_to_count = {}
        for result_set_name in sorted_result_set_names:
            image_count = image_counts[result_set_name]
            result_set_name_to_count[result_set_name] = image_count
        sorted_result_set_names = sorted(sorted_result_set_names,
                                         key=lambda x: result_set_name_to_count[x],
                                         reverse=True)

        for result_set_name in sorted_result_set_names:

            # Don't print classification classes here; we'll do that later with a slightly
            # different structure
            if has_classification_info and result_set_name.lower().startswith('class_'):
                continue

            filename = result_set_name + '.html'
            label = result_set_name_to_friendly_name(result_set_name)
            image_count = image_counts[result_set_name]

            # Don't include line items for empty multi-category pages
            if image_count == 0 and \
                detection_categories_to_category_count[result_set_name] > 1:
                    continue

            if total_images == 0:
                image_fraction = -1
            else:
                image_fraction = image_count / total_images

            # Write the line item for this category, including a link only if the
            # category is non-empty
            if image_count == 0:
                index_page += '{} ({}, {:.1%})<br/>\n'.format(
                    label,image_count,image_fraction)
            else:
                index_page += '<a href="{}">{}</a> ({}, {:.1%})<br/>\n'.format(
                    filename,label,image_count,image_fraction)

        # ...for each result set

        index_page += '</div>\n'

        # If classification information is present and we're supposed to create
        # a summary of classifications, we'll put it here
        category_count_footer = None

        if has_classification_info:

            index_page += '<h3>Species classification results</h3>'
            index_page += '<p>The same image might appear under multiple classes ' + \
                'if multiple species were detected.</p>\n'
            index_page += '<p>Classifications with confidence less than {:.1%} confidence are considered "unreliable".</p>\n'.format(
                options.classification_confidence_threshold)
            index_page += '<div class="contentdiv">\n'

            # Add links to all available classes
            class_names = sorted(classification_categories.values())
            if 'class_unreliable' in images_html.keys():
                class_names.append('unreliable')

            if options.sort_classification_results_by_count:
                class_name_to_count = {}
                for cname in class_names:
                    ccount = len(images_html['class_{}'.format(cname)])
                    class_name_to_count[cname] = ccount
                class_names = sorted(class_names,key=lambda x: class_name_to_count[x],reverse=True)

            if options.category_name_to_sort_weight is not None:
                category_name_to_sort_weight = {}
            else:
                category_name_to_sort_weight = options.category_name_to_sort_weight
                for category_name in category_name_to_sort_weight:
                    assert isinstance(category_name_to_sort_weight[category_name],int), \
                        'Illegal sort weight {} for category {}'.format(
                            category_name_to_sort_weight[category_name],category_name)

            # Figure out whether we need to do any grouping of categories
            # while we print results

            sort_weight_to_class_names = defaultdict(list)

            # Loop over class names in the already-sorted order, which will be
            # preserved within each weight group
            for class_name in class_names:
                if class_name in options.category_name_to_sort_weight:
                    weight = options.category_name_to_sort_weight[class_name]
                    sort_weight_to_class_names[weight].append(class_name)
                else:
                    # The default weight is zero
                    sort_weight_to_class_names[0].append(class_name)

            sort_weight_to_class_names = sort_dictionary_by_key(sort_weight_to_class_names)
            category_names_printed = set()

            for i_weight,sort_weight in enumerate(sort_weight_to_class_names):

                class_names_this_weight = sort_weight_to_class_names[sort_weight]

                # cname = class_names_this_weight[0]
                for cname in class_names_this_weight:

                    # Don't print multiple links when multiple category IDs have the same name
                    if cname in category_names_printed:
                        continue

                    category_names_printed.add(cname)
                    ccount = len(images_html['class_{}'.format(cname)])
                    if ccount > 0:
                        html_filename_base = clean_filename('class_{}'.format(cname),
                                                            replace_whitespace='_')
                        cname_print_string = cname.lower()
                        if len(cname_print_string) == 0:
                            cname_print_string = '[no name available]'
                        index_page += '<a href="{}.html">{}</a> ({})<br/>\n'.format(
                            html_filename_base, cname_print_string, ccount)

                # ...for every category in this sort weight group

                # Print a line break between sort weight groups
                if i_weight != (len(sort_weight_to_class_names)-1):
                    index_page += '<br/>\n'

            # ...for every sort weight group

            index_page += '</div>\n'

            if options.include_classification_category_report:

                # TODO: it's only for silly historical reasons that we re-read
                # the input file in this case; because this module has used Pandas
                # forever, we're not currently carrying the json representation around,
                # only the Pandas representation.

                print('Generating classification category report')

                d = load_md_or_speciesnet_file(options.md_results_file)

                category_name_to_count = {}

                include_category_descriptions_with_global_counts = \
                    options.include_category_descriptions_with_global_counts
                if 'classification_category_descriptions' not in d:
                    include_category_descriptions_with_global_counts = False
                else:
                    category_id_to_name = d['classification_categories']
                    category_name_to_id = invert_dictionary(category_id_to_name)
                    category_id_to_description = d['classification_category_descriptions']

                # im = d['images'][0]
                for im in d['images']:
                    if 'detections' in im and im['detections'] is not None:
                        for det in im['detections']:
                            if ('classifications' in det) and (len(det['classifications']) > 0):
                                class_id = det['classifications'][0][0]
                                class_conf = det['classifications'][0][1]
                                if class_conf < options.classification_confidence_threshold:
                                    continue
                                category_name = d['classification_categories'][class_id]
                                if category_name not in category_name_to_count:
                                    category_name_to_count[category_name] = 1
                                else:
                                    category_name_to_count[category_name] = \
                                        category_name_to_count[category_name] + 1
                        # ...for each detection
                    # ...if this image has detections
                # ...for each image

                category_name_to_count = sort_dictionary_by_value(
                    category_name_to_count,reverse=True)

                category_count_footer = ''
                category_count_footer += '<br/>\n'
                category_count_footer += \
                    '<h3>Category counts (for the whole dataset, not just the sample used for this page)</h3>\n'
                category_count_footer += '<div class="contentdiv">\n'

                for category_name in category_name_to_count.keys():
                    count = category_name_to_count[category_name]
                    category_count_html = '{}: {}'.format(category_name,count)
                    if include_category_descriptions_with_global_counts:
                        category_id = category_name_to_id[category_name]
                        category_description = category_id_to_description[category_id]
                        category_count_html += ' ({})'.format(category_description)
                    category_count_html += '<br/>\n'
                    category_count_footer += category_count_html

                category_count_footer += '</div>\n'

            # ...if we're generating a classification category report

        # ...if classification info is present

        if category_count_footer is not None:
            index_page += category_count_footer + '\n'

        # Write custom footer if it was provided
        if (options.footer_text is not None) and (len(options.footer_text) > 0):
            index_page += options.footer_text + '\n'

        # Close open html tags
        index_page += '\n</body></html>\n'

        output_html_file = os.path.join(output_dir, 'index.html')
        with open(output_html_file, 'w',
                  encoding=options.output_html_encoding) as f:
            f.write(index_page)

        print('Finished writing html to {}'.format(output_html_file))

    # ...if we do/don't have ground truth

    ppresults.output_html_file = output_html_file
    return ppresults

# ...process_batch_results


#%% Interactive driver(s)

if False:

    #%%

    base_dir = r'g:\temp'
    options = PostProcessingOptions()
    options.image_base_dir = base_dir
    options.output_dir = os.path.join(base_dir, 'preview')
    options.md_results_file = os.path.join(base_dir, 'results.json')
    options.confidence_threshold = {'person':0.5,'animal':0.5,'vehicle':0.01}
    options.include_almost_detections = True
    options.almost_detection_confidence_threshold = 0.001

    ppresults = process_batch_results(options)
    # from megadetector.utils.path_utils import open_file; open_file(ppresults.output_html_file)


#%% Command-line driver

def main(): # noqa

    options = PostProcessingOptions()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        'md_results_file',
        help='path to .json file containing MegaDetector results')
    parser.add_argument(
        'output_dir',
        help='base directory for output')
    parser.add_argument(
        '--image_base_dir', default=options.image_base_dir,
        help='base directory for images (optional, can compute statistics '
             'without images)')
    parser.add_argument(
        '--ground_truth_json_file', default=options.ground_truth_json_file,
        help='ground truth labels (optional, can render detections without '
             'ground truth), in the COCO Camera Traps format')
    parser.add_argument(
        '--confidence_threshold', type=float,
        default=options.confidence_threshold,
        help='Confidence threshold for statistics and visualization')
    parser.add_argument(
        '--almost_detection_confidence_threshold', type=float,
        default=options.almost_detection_confidence_threshold,
        help='Almost-detection confidence threshold for statistics and visualization')
    parser.add_argument(
        '--target_recall', type=float, default=options.target_recall,
        help='Target recall (for statistics only)')
    parser.add_argument(
        '--num_images_to_sample', type=int,
        default=options.num_images_to_sample,
        help='number of images to visualize, -1 for all images (default: 500)')
    parser.add_argument(
        '--viz_target_width', type=int, default=options.viz_target_width,
        help='Output image width')
    parser.add_argument(
        '--include_almost_detections', action='store_true',
        help='Include a separate category for images just above a second confidence threshold')
    parser.add_argument(
        '--html_sort_order', type=str, default='filename',
        help='Sort order for output pages, should be one of [filename,confidence,random] (defaults to filename)')
    parser.add_argument(
        '--sort_by_confidence', action='store_true',
        help='Sort output in decreasing order by confidence (defaults to sorting by filename)')
    parser.add_argument(
        '--n_cores', type=int, default=1,
        help='Number of threads to use for rendering (default: 1)')
    parser.add_argument(
        '--parallelize_rendering_with_processes',
        action='store_true',
        help='Should we use processes (instead of threads) for parallelization?')
    parser.add_argument(
        '--no_separate_detections_by_category',
        action='store_true',
        help='Collapse all categories into just "detections" and "non-detections"')
    parser.add_argument(
        '--open_output_file',
        action='store_true',
        help='Open the HTML output file when finished')
    parser.add_argument(
        '--max_figures_per_html_file',
        type=int, default=None,
        help='Maximum number of images to put on a single HTML page')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    if args.n_cores != 1:
        assert (args.n_cores > 1), 'Illegal number of cores: {}'.format(args.n_cores)
        if args.parallelize_rendering_with_processes:
            args.parallelize_rendering_with_threads = False
        args.parallelize_rendering = True
        args.parallelize_rendering_n_cores = args.n_cores

    args_to_object(args, options)

    if args.no_separate_detections_by_category:
        options.separate_detections_by_category = False

    ppresults = process_batch_results(options)

    if options.open_output_file:
        path_utils.open_file(ppresults.output_html_file)

if __name__ == '__main__':
    main()



"""

visualize_db.py

Outputs an HTML page visualizing annotations (class labels and/or bounding boxes)
on a sample of images in a database in the COCO Camera Traps format.

"""

#%% Imports

import argparse
import inspect
import random
import json
import math
import os
import sys
import time

import humanfriendly

from itertools import compress
from collections import defaultdict
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from copy import deepcopy
from tqdm import tqdm

from megadetector.utils.write_html_image_list import write_html_image_list
from megadetector.utils.path_utils import clean_filename
from megadetector.utils.ct_utils import is_empty
from megadetector.data_management.cct_json_utils import IndexedJsonDb
from megadetector.visualization import visualization_utils as vis_utils


#%% Settings

class DbVizOptions:
    """
    Parameters controlling the behavior of visualize_db().
    """

    def __init__(self):

        #: Number of images to sample from the database, or None to visualize all images
        self.num_to_visualize = None

        #: Target size for rendering; set either dimension to -1 to preserve aspect ratio.
        #:
        #: If viz_size is None or (-1,-1), the original image size is used.
        self.viz_size = (1000, -1)

        #: HTML rendering options; see write_html_image_list for details
        #:
        #:The most relevant option one might want to set here is:
        #:
        #: html_options['maxFiguresPerHtmlFile']
        #:
        #: ...which can be used to paginate previews to a number of images that will load well
        #: in a browser (1000 is a reasonable limit).
        self.html_options = write_html_image_list()

        #: Whether to sort images by filename (True) or randomly (False)
        #:
        #: This is ignored if max_sequence_length is not None, in which case
        #: we always sort by sequence ID, then frame number.
        self.sort_by_filename = True

        #: Only show images that contain bounding boxes
        self.trim_to_images_with_bboxes = False

        #: Random seed to use for sampling images
        self.random_seed = 0

        #: Should we include Web search links for each category name?
        self.add_search_links = False

        #: Should each thumbnail image link back to the original image?
        self.include_image_links = False

        #: Should there be a text link back to each original image?
        self.include_filename_links = False

        #: Line width in pixels
        self.box_thickness = 4

        #: Number of pixels to expand each bounding box
        self.box_expansion = 0

        #: Only include images that contain annotations with these class names (not IDs) (list)
        #:
        #: Mutually exclusive with classes_to_exclude
        self.classes_to_include = None

        #: Exclude images that contain annotations with these class names (not IDs) (list)
        #:
        #: Mutually exclusive with classes_to_include
        self.classes_to_exclude = None

        #: Special tag used to say "show me all images with multiple categories"
        #:
        #: :meta private:
        self.multiple_categories_tag = '*multiple*'

        #: Parallelize rendering across multiple workers
        self.parallelize_rendering = False

        #: In theory, whether to parallelize with threads (True) or processes (False), but
        #: process-based parallelization in this function is currently unsupported
        self.parallelize_rendering_with_threads = True

        #: Number of workers to use for parallelization; ignored if parallelize_rendering
        #: is False
        self.parallelize_rendering_n_cores = 16

        #: Should we show absolute (True) or relative (False) paths for each image?
        self.show_full_paths = False

        #: List of additional fields in the image struct that we should print in image headers
        self.extra_image_fields_to_print = None

        #: List of additional fields in the annotation struct that we should print in image headers
        self.extra_annotation_fields_to_print = None

        #: Set to False to skip existing images
        self.force_rendering = True

        #: Enable additionald debug console output
        self.verbose = False

        #: COCO files used for evaluation may contain confidence scores, this
        #: determines the field name used for confidence scores
        self.confidence_field_name = 'score'

        #: Optionally apply a confidence threshold; this requires that [confidence_field_name]
        #: be present in all detections.
        self.confidence_threshold = None

        #: Custom mapping from category IDs to labels, replacing what's in the .json file
        self.custom_category_mapping = None

        #: JPEG quality to use for saving images (None for Pillow default)
        self.quality = None

        #: List of PIL color names, which will be indexed by category IDs, or None
        #: to use the default color map.
        #:
        #: For example: ['AliceBlue', 'Red', 'RoyalBlue', 'Gold', 'Chartreuse']
        self.colormap = None

        #: Should we create separate pages for each category (within the sampled set)?
        #:
        #: Images with multiple categories will be included in all relevant pages.
        self.create_category_pages = False

        #: If this is None, we just sample images, and show images.  If this is
        #: not None, we sample images, but we also show the other images in the sequences
        #: containing our sampled images.  If this is <=0, there is no limit on the
        #: number of images we'll show per sequences.  If this is >0, we will cap the number
        #: of images shown per sequence; no guarantee is made about which images will
        #: be selected in that case.  This only impacts the number of images added as
        #: "sequence friends" of images that get sampled.
        self.max_sequence_length = None


#%% Core functions

def visualize_db(db_path, output_dir, image_base_dir, options=None):
    """
    Writes images and html to output_dir to visualize the images and annotations in a
    COCO-formatted .json file.

    Args:
        db_path (str or dict): the .json filename to load, or a previously-loaded database
        output_dir (str): the folder to which we should write annotated images
        image_base_dir (str): the folder where the images live; filenames in [db_path] should
            be relative to this folder.
        options (DbVizOptions, optional): See DbVizOptions for details

    Returns:
        tuple: A length-two tuple containing (the html filename) and (the loaded database).
    """

    if options is None:
        options = DbVizOptions()

    # Consistency checking for fields with specific format requirements

    # These should be a lists, but if someone specifies a string, do a reasonable thing
    if isinstance(options.extra_image_fields_to_print,str):
        options.extra_image_fields_to_print = [options.extra_image_fields_to_print]
    if isinstance(options.extra_annotation_fields_to_print,str):
        options.extra_annotation_fields_to_print = [options.extra_annotation_fields_to_print]

    if not options.parallelize_rendering_with_threads:
        print('Warning: process-based parallelization is not yet supported by visualize_db')
        options.parallelize_rendering_with_threads = True

    if image_base_dir.startswith('http'):
        if not image_base_dir.endswith('/'):
            image_base_dir += '/'
    else:
        assert(os.path.isdir(image_base_dir))

    os.makedirs(os.path.join(output_dir, 'rendered_images'), exist_ok=True)

    if isinstance(db_path,str):
        assert(os.path.isfile(db_path))
        print('Loading database from {}...'.format(db_path))
        image_db = json.load(open(db_path))
        print('...done, loaded {} images'.format(len(image_db['images'])))
    elif isinstance(db_path,dict):
        print('Using previously-loaded DB')
        image_db = db_path
    else:
        raise ValueError('Illegal dictionary or filename')

    annotations = image_db['annotations']
    images = image_db['images']
    categories = image_db['categories']

    # Optionally remove all images without bounding boxes, *before* sampling
    if options.trim_to_images_with_bboxes:

        b_has_bbox = [False] * len(annotations)
        for i_ann,ann in enumerate(annotations):
            if 'bbox' in ann or 'bbox_relative' in ann:
                if 'bbox' in ann:
                    assert isinstance(ann['bbox'],list)
                else:
                    assert isinstance(ann['bbox_relative'],list)
                b_has_bbox[i_ann] = True
        annotations_with_boxes = list(compress(annotations, b_has_bbox))

        image_ids_with_boxes = [x['image_id'] for x in annotations_with_boxes]
        image_ids_with_boxes = set(image_ids_with_boxes)

        image_has_box = [False] * len(images)
        for i_image,image in enumerate(images):
            image_id = image['id']
            if image_id in image_ids_with_boxes:
                image_has_box[i_image] = True
        images_with_bboxes = list(compress(images, image_has_box))
        images = images_with_bboxes

    # Optionally include/remove images with specific labels, *before* sampling

    assert (not ((options.classes_to_exclude is not None) and \
                 (options.classes_to_include is not None))), \
                 'Cannot specify an inclusion and exclusion list'

    if options.classes_to_exclude is not None:
        assert isinstance(options.classes_to_exclude,list), \
            'If supplied, classes_to_exclude should be a list'

    if options.classes_to_include is not None:
        assert isinstance(options.classes_to_include,list), \
            'If supplied, classes_to_include should be a list'

    if (options.classes_to_exclude is not None) or (options.classes_to_include is not None):

        print('Indexing database')
        indexed_db = IndexedJsonDb(image_db)
        b_valid_class = [True] * len(images)
        for i_image,image in enumerate(images):
            classes = indexed_db.get_classes_for_image(image)
            if options.classes_to_exclude is not None:
                for excluded_class in options.classes_to_exclude:
                    if excluded_class in classes:
                       b_valid_class[i_image] = False
                       break
            elif options.classes_to_include is not None:
                b_valid_class[i_image] = False
                if options.multiple_categories_tag in options.classes_to_include:
                    if len(classes) > 1:
                        b_valid_class[i_image] = True
                if not b_valid_class[i_image]:
                    for c in classes:
                        if c in options.classes_to_include:
                            b_valid_class[i_image] = True
                            break
            else:
                raise ValueError('Illegal include/exclude combination')

        images_with_valid_classes = list(compress(images, b_valid_class))
        images = images_with_valid_classes

    # ...if we need to include/exclude categories

    # Construct label map
    label_map = {}
    for cat in categories:
        label_map[int(cat['id'])] = cat['name']

    # Map sequence IDs to images if necessary
    sequence_id_to_images = defaultdict(list)

    if options.max_sequence_length is not None:
        for im in images:
            if 'seq_id' in im:
                sequence_id_to_images[im['seq_id']].append(im)

    # Take a sample of images
    if options.num_to_visualize is not None:
        if options.num_to_visualize > len(images):
            print('Warning: asked to visualize {} images, but only {} are available, keeping them all'.\
                  format(options.num_to_visualize,len(images)))
        else:
            random.seed(options.random_seed)
            images = random.sample(images,options.num_to_visualize)

    # List of dicts containing HTML image information suitable for passing to
    # write_html_image_list
    images_html = []

    # Set of dicts representing inputs to _render_image_info, with fields:
    #
    # bboxes, box_classes, tags, img_path, output_file_name, boxes_are_normalies
    rendering_info = []

    image_id_to_annotations = defaultdict(list)

    for ann in annotations:
        image_id_to_annotations[ann['image_id']].append(ann)

    # Do we need to find all the images that are in the same
    # sequence as the sampled images?
    if options.max_sequence_length is not None:

        print('Preparing sequence friends')

        seq_id_to_images_rendered = defaultdict(int)

        # Don't add images that are already in the list
        image_ids_already_sampled = set()
        for im in images:
            image_ids_already_sampled.add(im['id'])
            if 'seq_id' in im:
                seq_id_to_images_rendered[im['seq_id']] += 1

        images_to_add = []

        # For every image we've decided to render, find all her
        # sequence friends
        for im in images:

            if 'seq_id' not in im:
                continue

            seq_id = im['seq_id']

            # This is *all* the images in this sequence, regardless of
            # whether we've already added them
            images_this_sequence = sequence_id_to_images[seq_id]

            for candidate_im in images_this_sequence:

                # Have we already hit the maximum number of images for this sequence?
                if (options.max_sequence_length > 0) and \
                   (seq_id_to_images_rendered[seq_id] >= options.max_sequence_length):
                    # Don't assert this, we may have more than max_sequence_length images
                    # if we just happened to sample that way, we just won't add more.
                    # assert seq_id_to_images_rendered[seq_id] == options.max_sequence_length
                    if options.verbose:
                        print('Already rendered {} images from sequence {}'.format(
                            options.max_sequence_length,seq_id))
                    break

                assert candidate_im['seq_id'] == seq_id

                # Add this image if necessary
                if candidate_im['id'] not in image_ids_already_sampled:

                    images_to_add.append(candidate_im)
                    image_ids_already_sampled.add(candidate_im['id'])
                    seq_id_to_images_rendered[seq_id] += 1

            # ...for each sequence friend of the current image

        # ...for each image in the sampled set

        print('Adding {} new images ({} initially) as sequence friends'.format(
            len(images_to_add),len(images)))
        images.extend(images_to_add)

        # Double-check that we didn't create duplicates
        all_image_ids = [im['id'] for im in images]
        assert len(all_image_ids) == len(set(all_image_ids))

    # ...if we need to add sequence friends for every sampled image

    print('Preparing rendering list')

    for i_image,img in tqdm(enumerate(images),total=len(images)):

        img_id = img['id']
        assert img_id is not None

        img_relative_path = img['file_name']

        if image_base_dir.startswith('http'):
            img_path = image_base_dir + img_relative_path
        else:
            img_path = os.path.join(image_base_dir,img_relative_path).replace('\\','/')

        if img_id in image_id_to_annotations:
            annotations_this_image = image_id_to_annotations[img_id]
        else:
            annotations_this_image = []

        bboxes = []
        box_classes = []
        box_score_strings = []

        # All the class labels we've seen for this image (with or without bboxes)
        image_categories = set()

        extra_annotation_field_string = ''
        annotation_level_for_image = ''

        # Did this image come with already-normalized bounding boxes?
        boxes_are_normalized = None

        # Iterate over annotations for this image
        for i_ann,ann in enumerate(annotations_this_image):

            if options.extra_annotation_fields_to_print is not None:
                field_names = ann.keys()
                for field_name in field_names:
                    if field_name in options.extra_annotation_fields_to_print:
                        field_value = ann[field_name]
                        if not is_empty(field_value):
                            extra_annotation_field_string += ' ({}:{})'.format(
                                field_name,field_value)

            if options.confidence_threshold is not None:
                assert options.confidence_field_name in ann, \
                    'Error: confidence thresholding requested, ' + \
                        'but at least one annotation does not have the {} field'.format(
                            options.confidence_field_name)
                if ann[options.confidence_field_name] < options.confidence_threshold:
                    continue

            if 'sequence_level_annotation' in ann:
                b_sequence_level_annotation = ann['sequence_level_annotation']
                if b_sequence_level_annotation:
                    annotation_level = 'sequence'
                else:
                    annotation_level = 'image'
                if annotation_level_for_image == '':
                    annotation_level_for_image = annotation_level
                elif annotation_level_for_image != annotation_level:
                    annotation_level_for_image = 'mixed'

            category_id = ann['category_id']
            category_name = label_map[category_id]
            if options.add_search_links:
                category_name = category_name.replace('"','')
                category_name = '<a href="https://www.google.com/search?tbm=isch&q={}">{}</a>'.format(
                    category_name,category_name)

            image_categories.add(category_name)

            assert not ('bbox' in ann and 'bbox_relative' in ann), \
                "An annotation can't have both an absolute and a relative bounding box"

            box_field = 'bbox'
            if 'bbox_relative' in ann:
                box_field = 'bbox_relative'
                assert (boxes_are_normalized is None) or (boxes_are_normalized), \
                    "An image can't have both absolute and relative bounding boxes"
                boxes_are_normalized = True
            elif 'bbox' in ann:
                assert (boxes_are_normalized is None) or (not boxes_are_normalized), \
                    "An image can't have both absolute and relative bounding boxes"
                boxes_are_normalized = False

            if box_field in ann:
                bbox = ann[box_field]
                if isinstance(bbox,float):
                    assert math.isnan(bbox), "I shouldn't see a bbox that's neither a box nor NaN"
                    continue
                bboxes.append(bbox)
                box_classes.append(ann['category_id'])

                box_score_string = ''
                if options.confidence_field_name is not None and \
                   options.confidence_field_name in ann:
                       score = ann[options.confidence_field_name]
                       box_score_string = '({}%)'.format(round(100 * score))
                box_score_strings.append(box_score_string)

        # ...for each of this image's annotations

        image_classes = ', '.join(image_categories)

        img_id_string = str(img_id).lower()
        file_name = '{}_gt.jpg'.format(os.path.splitext(img_id_string)[0])

        # Replace characters that muck up image links, including flattening file
        # separators.
        illegal_characters = ['/','\\',':','\t','#',' ','%']
        for c in illegal_characters:
            file_name = file_name.replace(c,'~')

        rendering_info_this_image = {'bboxes':bboxes,
                                     'box_classes':box_classes,
                                     'tags':box_score_strings,
                                     'img_path':img_path,
                                     'output_file_name':file_name,
                                     'boxes_are_normalized':boxes_are_normalized}
        rendering_info.append(rendering_info_this_image)

        label_level_string = ''
        if len(annotation_level_for_image) > 0:
            label_level_string = ' (annotation level: {})'.format(annotation_level_for_image)

        if 'frame_num' in img and 'seq_num_frames' in img:
            frame_string = ' frame: {} of {},'.format(img['frame_num'],img['seq_num_frames'])
        elif 'frame_num' in img:
            frame_string = ' frame: {},'.format(img['frame_num'])
        else:
            frame_string = ''

        if options.show_full_paths:
            filename_text = img_path
        else:
            filename_text = img_relative_path
        if options.include_filename_links:
            filename_text = '<a href="{}">{}</a>'.format(img_path,filename_text)

        flag_string = ''

        if ('flags' in img) and (not is_empty(img['flags'])):
            flag_string = ', flags: {}'.format(str(img['flags']))

        extra_field_string = ''

        if options.extra_image_fields_to_print is not None:
            for field_name in options.extra_image_fields_to_print:
                if field_name in img:
                    # Always include a leading comma; this either separates us from the
                    # previous field in [extra_fields_to_print] or from the rest of the string
                    extra_field_string += ', {}: {}'.format(
                        field_name,str(img[field_name]))

        # We're adding html for an image before we render it, so it's possible this image will
        # fail to render.  For applications where this script is being used to debug a database
        # (the common case), this is useful behavior, because we still see a record of this
        # item in the HTML output.  For other applications, it's annoying to have broken images.
        image_dict = \
        {
            'filename': '{}/{}'.format('rendered_images', file_name),
            'title': '{}<br/>{}, num boxes: {},{} class labels: {}{}{}{}{}'.format(
                filename_text, img_id, len(bboxes), frame_string, image_classes,
                label_level_string, flag_string, extra_field_string, extra_annotation_field_string),
            'textStyle': 'font-family:verdana,arial,calibri;font-size:80%;' + \
                'text-align:left;margin-top:20;margin-bottom:5;',
            'imageStyle':'max-width:95%;',
            'image_categories': image_categories
        }

        # Make it clear which images start a new sequence
        if (options.max_sequence_length is not None) and \
           ('frame_num' in img) and ('seq_id' in img) and \
           (img['frame_num'] is not None) and (img['frame_num'] in (0,-1)):
            sequence_header = '<span style="font-weight:bold;font-size:120%">Sequence {}</span><br/><br/>'.format(
                img['seq_id'])
            image_dict['title'] = sequence_header + image_dict['title']

        if options.include_image_links:
            image_dict['linkTarget'] = img_path

        for field_name in ('seq_id','frame_num'):
            if field_name in img:
                image_dict[field_name] = img[field_name]

        images_html.append(image_dict)

    # ...for each sampled image

    def _render_image_info(rendering_info):
        """
        Render one image.

        Args:
            rendering_info (dict): a dict with fields:
                - img_path: the absolute path (or URL) to the input image
                - bboxes: a list of bbox dicts to render (or [])
                - bbox_classes: a list of category IDs the same length as [bboxes]
                - output_file_name: the relative output path to which we should render
                - tags (optional): a list of additional text tags to include with each box

        Returns:
            bool: True if rendering was successful, else False
        """

        img_path = rendering_info['img_path']
        bboxes = rendering_info['bboxes']
        bbox_classes = rendering_info['box_classes']
        boxes_are_normalized = rendering_info['boxes_are_normalized']
        bbox_tags = None
        if 'tags' in rendering_info:
            bbox_tags = rendering_info['tags']
        output_file_name = rendering_info['output_file_name']
        output_full_path = os.path.join(output_dir, 'rendered_images', output_file_name)

        if (os.path.isfile(output_full_path)) and (not options.force_rendering):
            if options.verbose:
                print('Skipping existing image {}'.format(output_full_path))
            return True

        if not img_path.startswith('http'):
            if not os.path.exists(img_path):
                print('Image {} cannot be found'.format(img_path))
                return False

        try:
            original_image = vis_utils.open_image(img_path)
            original_size = original_image.size
            if (options.viz_size is None) or \
                (options.viz_size[0] == -1 and options.viz_size[1] == -1):
                image = original_image
            else:
                image = vis_utils.resize_image(original_image,
                                               options.viz_size[0],
                                               options.viz_size[1],
                                               no_enlarge_width=True)
        except Exception as e:
            print('Image {} failed to open, error: {}'.format(img_path, e))
            return False

        rendering_label_map = label_map
        if options.custom_category_mapping is not None:
            rendering_label_map = options.custom_category_mapping

        vis_utils.render_db_bounding_boxes(boxes=bboxes,
                                           classes=bbox_classes,
                                           image=image,
                                           original_size=original_size,
                                           label_map=rendering_label_map,
                                           thickness=options.box_thickness,
                                           expansion=options.box_expansion,
                                           tags=bbox_tags,
                                           boxes_are_normalized=boxes_are_normalized,
                                           colormap=options.colormap)

        if options.quality is None:
            image.save(output_full_path)
        else:
            image.save(output_full_path,quality=options.quality)

        return True

    # ...def _render_image_info(...)

    print('Rendering images')
    start_time = time.time()

    if options.parallelize_rendering:

        if options.parallelize_rendering_with_threads:
            worker_string = 'threads'
        else:
            worker_string = 'processes'

        pool = None
        try:
            if options.parallelize_rendering_n_cores is None:
                if options.parallelize_rendering_with_threads:
                    pool = ThreadPool()
                else:
                    pool = Pool()
            else:
                if options.parallelize_rendering_with_threads:
                    pool = ThreadPool(options.parallelize_rendering_n_cores)
                else:
                    pool = Pool(options.parallelize_rendering_n_cores)
                print('Rendering images with {} {}'.format(options.parallelize_rendering_n_cores,
                                                           worker_string))
            rendering_success = list(tqdm(pool.imap(_render_image_info, rendering_info),
                                     total=len(rendering_info)))
        finally:
            if pool is not None:
                pool.close()
                pool.join()
                print("Pool closed and joined for DB visualization")

    else:

        rendering_success = []
        for file_info in tqdm(rendering_info):
            rendering_success.append(_render_image_info(file_info))

    elapsed = time.time() - start_time

    print('Rendered {} images in {} ({} successful)'.format(
        len(rendering_info),humanfriendly.format_timespan(elapsed),sum(rendering_success)))

    # If we added every sampled image's sequence friends, we sort by sequence ID,
    # then frame number
    if options.max_sequence_length is not None:

        # Add fields we need for sorting by sequence
        for im in images_html:
            if 'seq_id' not in im:
                im['seq_id'] = 'no_sequence_id'
            if 'frame_num' not in im:
                im['frame_num'] = -1

        # Sort the list of dicts "images_html" by the
        # field "seq_id", then "frame_num", then "filename"
        images_html = sorted(images_html,
                             key=lambda x: (x['seq_id'], x['frame_num'], x['filename']))

    else:

        if options.sort_by_filename:
            images_html = sorted(images_html, key=lambda x: x['filename'])
        else:
            random.shuffle(images_html)

    # ...if we need to sort using sequence information

    html_output_file = None

    if options.create_category_pages:

        all_categories = set()
        for im in images_html:
            categories_this_image = im['image_categories']
            for category_name in categories_this_image:
                all_categories.add(category_name)

        all_categories = sorted(list(all_categories))

        # Create a special category for images with no annotations
        no_category_token = 'no_categories'
        all_categories.append(no_category_token)

        category_name_to_relative_filename = {}
        category_name_to_count = {}

        for category_name in all_categories:

            category_name_clean = clean_filename(category_name,
                                                 force_lower=True,
                                                 replace_whitespace='_')
            category_filename_relative = category_name_clean + '.html'
            category_name_to_relative_filename[category_name] = category_filename_relative
            category_filename_abs = os.path.join(output_dir,category_filename_relative)

            # Find images that should go in this category page
            images_this_category = []

            for im in images_html:

                include_this_image = False
                if category_name == no_category_token:
                    include_this_image = (len(im['image_categories']) == 0)
                else:
                    include_this_image = category_name in im['image_categories']
                if include_this_image:
                    images_this_category.append(im)

            # ...for each image

            category_name_to_count[category_name] = len(images_this_category)

            title_string = '<h1>Sample annotations for category {}</h1>'.format(
                category_name)

            html_options = deepcopy(options.html_options)
            html_options['headerHtml'] = title_string

            write_html_image_list(
                filename=category_filename_abs,
                images=images_this_category,
                options=html_options)

        # ...for each category

        html_output_file = os.path.join(output_dir, 'index.html')

        if isinstance(db_path,str):
            title_string = '<h1>Sample annotations from {}</h1>'.format(db_path)
        else:
            title_string = '<h1>Sample annotations</h1>'

        with open(html_output_file,'w') as f:
            f.write('<html><head>{}</head><body>\n'.format(title_string))
            for category_name in category_name_to_relative_filename:
                s = '<p style="padding:0px;margin:0px;margin-left:15px;text-align:left;'
                s += 'font-family:''segoe ui'',calibri,arial;font-size:100%;text-decoration:none;font-weight:normal;">'
                f.write(s)
                f.write('<a href="{}">{}</a> ({} images)</p>\n'.format(
                    category_name_to_relative_filename[category_name],
                    category_name,
                    category_name_to_count[category_name]
                ))
            f.write('</body></html>\n')

    else:

        html_output_file = os.path.join(output_dir, 'index.html')

        html_options = options.html_options
        if isinstance(db_path,str):
            html_options['headerHtml'] = '<h1>Sample annotations from {}</h1>'.format(db_path)
        else:
            html_options['headerHtml'] = '<h1>Sample annotations</h1>'

        write_html_image_list(
            filename=html_output_file,
            images=images_html,
            options=html_options)

        print('Visualized {} images, wrote results to {}'.format(len(images_html),html_output_file))

    return html_output_file,image_db

# ...def visualize_db(...)


#%% Command-line driver

# Copy all fields from a Namespace (i.e., the output from parse_args) to an object.
#
# Skips fields starting with _.  Does not check existence in the target object.
def _args_to_object(args, obj):

    for n, v in inspect.getmembers(args):
        if not n.startswith('_'):
            setattr(obj, n, v)

def main():
    """
    Command-line driver for visualize_db
    """

    parser = argparse.ArgumentParser()
    parser.add_argument('db_path', action='store', type=str,
                        help='.json file to visualize')
    parser.add_argument('output_dir', action='store', type=str,
                        help='Output directory for html and rendered images')
    parser.add_argument('image_base_dir', action='store', type=str,
                        help='Base directory (or URL) for input images')

    parser.add_argument('--num_to_visualize', action='store', type=int, default=None,
                        help='Number of images to visualize (randomly drawn) (defaults to all)')
    parser.add_argument('--random_sort', action='store_true',
                        help='Sort randomly (rather than by filename) in output html')
    parser.add_argument('--trim_to_images_with_bboxes', action='store_true',
                        help='Only include images with bounding boxes (defaults to false)')
    parser.add_argument('--random_seed', action='store', type=int, default=None,
                        help='Random seed for image selection')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    # Convert to an options object
    options = DbVizOptions()
    _args_to_object(args, options)
    if options.random_sort:
        options.sort_by_filename = False

    visualize_db(options.db_path,options.output_dir,options.image_base_dir,options)

if __name__ == '__main__':
    main()



"""

threshold_coco_dataset.py

Given a COCO-formatted dataset that stores confidence in the semi-standard "score"
field, remove annotations below a threshold.

"""

#%% Imports and constants

import json
import sys
import argparse

from tqdm import tqdm

from megadetector.utils.ct_utils import write_json


#%% Functions

def threshold_coco_dataset(input_filename,
                           confidence_threshold=0.0,
                           output_filename=None,
                           confidence_field='score',
                           missing_confidence_handling='error'):
    """
    Given a COCO-formatted dataset that stores confidence in the semi-standard "score"
    field, remove annotations below a threshold.


    Args:
        input_filename (str): the (input) COCO-formatted .json file containing annotations
        confidence_threshold (float, optional): discard annotations below this confidence value
        output_filename (str, optional): write the thresholded output here
        confidence_field (str, optional): the field within annotations that
            represents confidence values
        missing_confidence_handling: what to do if a confidence value is missing
            (should be 'error' or 'warning')

    Returns:
        dict: the thresholded COCO database
    """

    # Validate arguments

    assert missing_confidence_handling in ('error','warning'), \
        f'Illegal missing confidence handling {missing_confidence_handling}'

    # Read input data
    print('Reading file {}'.format(input_filename))
    with open(input_filename,'r') as f:
        d = json.load(f)

    annotations_to_keep = []
    annotations = d['annotations']

    for ann in tqdm(annotations):

        if confidence_field not in ann:
            ann_id = 'unknown'
            if 'id' in ann:
                ann_id = ann['id']
            s = 'annotation {} is missing field {}'.format(ann_id,confidence_field)
            if missing_confidence_handling == 'error':
                raise ValueError(s)
            else:
                assert missing_confidence_handling == 'warning'
                print('Warning: {}'.format(s))
                continue

        confidence = ann[confidence_field]
        if confidence >= confidence_threshold:
            annotations_to_keep.append(ann)

    # ...for each annotation

    print('Keeping {} of {} annotations'.format(
        len(annotations_to_keep),len(d['annotations'])))
    d['annotations'] = annotations_to_keep

    if output_filename is not None:
        write_json(output_filename,d)

    return d

# ...def threshold_coco_dataset(...)


#%% Command-line driver

def main():
    """
    Command-line driver for threshold_coco_dataset
    """

    parser = argparse.ArgumentParser(
        description='Threshold a COCO dataset, write the results to a new file'
    )
    parser.add_argument(
        'input_filename',
        type=str,
        help='Path to the input COCO .json file'
    )
    parser.add_argument(
        'output_filename',
        type=str,
        help='Path to the .json file where thresholded data will be saved'
    )
    parser.add_argument(
        'confidence_threshold',
        type=float,
        help='Confidence threshold'
    )
    parser.add_argument(
        '--confidence_field',
        type=str,
        default='score',
        help='Field to use for confidence values, default "score"'
    )
    parser.add_argument(
        '--missing_confidence_handling',
        type=str,
        default='error',
        choices=['error', 'warning'],
        help='Whether to error on annotations that are missing the confidence field'
    )

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    threshold_coco_dataset(
        input_filename=args.input_filename,
        confidence_threshold=args.confidence_threshold,
        output_filename=args.output_filename,
        confidence_field=args.confidence_field,
        missing_confidence_handling=args.missing_confidence_handling
    )


if __name__ == '__main__':
    main()



"""

wi_download_csv_to_coco.py

Converts a .csv file (or a folder of .csv files) from a Wildlife Insights project export to a
COCO camera traps .json file.

Currently assumes that common names are unique identifiers, which is convenient but unreliable.

"""

#%% Imports and constants

import os
import re

from tqdm import tqdm
from collections import defaultdict

from megadetector.utils.ct_utils import write_json
from megadetector.utils.ct_utils import is_empty
from megadetector.utils.ct_utils import sort_dictionary_by_value
from megadetector.utils.ct_utils import sort_list_of_dicts_by_key
from megadetector.utils.string_utils import is_int
from megadetector.utils.path_utils import find_images
from megadetector.utils.wi_platform_utils import read_images_from_download_bundle
from megadetector.utils.wi_platform_utils import read_sequences_from_download_bundle
from megadetector.utils.wi_platform_utils import url_to_relative_path

wi_extra_annotation_columns = \
    ('identified_by',
     'wi_taxon_id',
     'uncertainty',
     'number_of_objects',
     'group_size',
     'age',
     'sex',
     'animal_recognizable',
     'individual_id',
     'individual_animal_notes',
     'behavior',
     'highlighted',
     'markings')

# Omitted:
#
# is_blank
# filename
# cv_confidence
# license
# bounding_boxes

# Handled as part of the category:
#
# class
# order
# family
# genus
# species

# Handled as part of the image:
#
# timestamp
# image_id
# project_id
# deployment_id
# location

wi_extra_image_columns = ('project_id','deployment_id')

def _make_location_id(project_id,deployment_id):
    return 'project_' + str(project_id) + '_deployment_' + deployment_id

# The default category mappings lose some information about vehicles,
# but we are typically using this to perform accuracy comparisons for animals,
# and having lots of categories that co-occur with other categories (which
# vehicles do) makes analysis messier.
default_category_remappings = {
    # "blank" is handled specially below
    # 'blank':'empty',
    'homo species':'human',
    'no cv result':'unknown',
    'misfire':'blank',
    '.*human.*':'human',
    '.*vehicle.*':'vehicle',
    'truck':'vehicle',
    'atv':'vehicle'
}


#%% Main function

def wi_download_csv_to_coco(csv_file_in,
                            coco_file_out=None,
                            image_folder=None,
                            exclude_missing_images=False,
                            image_flattening='deployment',
                            verbose=True,
                            category_remappings=default_category_remappings,
                            blank_disagreement_handling='trust_label',
                            include_blanks=True):
    """
    Converts a .csv file (or folder of .csv files) from a Wildlife Insights project export
    to a COCO Camera Traps .json file.

    TODO: currently relies on uniqueness of common names, which is not guaranteed.  Prints
    warnings for non-unique common names.

    Args:
        csv_file_in (str): a downloaded .csv file we should convert to COCO, or a folder
            containing images...csv files.
        coco_file_out (str, optional): the .json file we should write; if [coco_file_out] is None,
            returns data, but doesn't write it
        image_folder (str, optional): the folder where images live, only relevant if
            [exclude_missing_images] is True
        exclude_missing_images (bool, optional): whether to exclude images not present
            in disk; if this is True, [image_folder] must be a valid folder.  This has no
            impact on blank images if "include_blanks" is False.
        image_flattening (str, optional): if 'none', relative paths will be stored
            as the entire URL for each image, other than gs://.  Can be 'guid' (just
            store [GUID].JPG) or 'deployment' (store as [deployment]/[GUID].JPG).
        verbose (bool, optional): enable additional debug console output
        category_remappings (dict, optional): str --> str dict that maps WI category
            names to output category names.  Regular expressions allowed in keys.
        blank_disagreement_handling (str, optional): what to do when the "common_name"
            field disagrees with the "is_blank" field; can be "trust_label" (default),
            "trust_is_blank", or "error
        include_blanks (bool, optional): whether to include blank images in the COCO
            file

    Returns:
        dict: COCO-formatted data, identical to what's written to [coco_file_out]
    """

    ##%% Validate inputs

    assert os.path.isfile(csv_file_in) or os.path.isdir(csv_file_in), \
        '{} does not exist'.format(csv_file_in)

    assert blank_disagreement_handling in ('trust_label','trust_is_blank','error'), \
        'Unknown blank disagreement handling value: {}'.format(
            blank_disagreement_handling)


    ##%% Read input files

    # read_images_from_download_bundle supports a folder or a single .csv file
    image_id_to_image_records = read_images_from_download_bundle(csv_file_in)

    assert image_id_to_image_records is not None, \
        'Failed to read images from {}'.format(csv_file_in)

    print('Read image records for {} unique image IDs'.format(
        len(image_id_to_image_records)))

    sequence_id_to_sequence_records = read_sequences_from_download_bundle(csv_file_in)

    sequence_id_to_image_ids = None

    # Is this a sequence-based project?
    if sequence_id_to_sequence_records is not None:

        print('Read sequence records for {} sequence IDs'.format(
            len(sequence_id_to_sequence_records)))

        # Group images into sequences

        sequence_id_to_image_ids = defaultdict(set)

        # image_id = next(iter(image_id_to_image_records))
        for image_id in image_id_to_image_records:

            records_this_image = image_id_to_image_records[image_id]

            for r in records_this_image:

                assert image_id == r['image_id']
                if 'sequence_id' not in r:
                    print('Warning: image {} does not have a sequence ID'.format(r['image_id']))
                    continue
                sequence_id_to_image_ids[r['sequence_id']].add(image_id)

            # ...for each record associated with this image ID

        # ...for each image ID

        # Create frame numbers and frame ordering

        # sequence_id = next(iter(sequence_id_to_image_ids))
        for sequence_id in sequence_id_to_image_ids:

            image_ids_this_sequence = sequence_id_to_image_ids[sequence_id]

            records_this_sequence = []

            for image_id in image_ids_this_sequence:

                records_this_image = image_id_to_image_records[image_id]
                # Choose a representative record for sorting
                r = records_this_image[0]
                # Timestamps are formatted as "2019-09-09 13:45:00"
                assert isinstance(r['timestamp'],str) and len(r['timestamp']) == 19
                records_this_sequence.append(r)

            sorted_records_this_sequence = \
                sort_list_of_dicts_by_key(records_this_sequence,'timestamp')

            # i_record = 0; r = sorted_records_this_sequence[i_record]
            for i_record,r in enumerate(sorted_records_this_sequence):

                r['frame_num'] = i_record
                r['seq_num_frames'] = len(sorted_records_this_sequence)
                image_id = r['image_id']

                # If there are multiple records for this image (typically indicating multiple
                # species), propagate that information to the other records
                records_this_image_id = image_id_to_image_records[image_id]

                # target_r = records_this_image_id[0]
                for target_r in records_this_image_id:

                    if r == target_r:
                        continue

                    assert r['timestamp'] == target_r['timestamp']
                    target_r['frame_num'] = i_record
                    target_r['seq_num_frames'] = len(sorted_records_this_sequence)

            # ...for each record in this sequence

        # ...for each sequence ID

    # ...if this is a sequence-based project


    #%% Create COCO dictionaries

    category_name_to_category = {}
    empty_category = {'name':'empty','id':0,'count':0,'taxonomy_string':''}
    category_name_to_category['empty'] = empty_category

    image_id_to_image = {}
    image_id_to_annotations = defaultdict(list)

    print('Converting records to COCO...')

    n_blanks_excluded = 0

    # image_id = next(iter(image_id_to_image_records))
    for image_id in tqdm(image_id_to_image_records.keys(),
                         total=len(image_id_to_image_records)):

        image_records_this_id = image_id_to_image_records[image_id]

        reference_record = image_records_this_id[0]

        url = reference_record['location']
        assert url.startswith('gs://')

        file_name = url_to_relative_path(url,image_flattening=image_flattening)

        location_id = _make_location_id(
            reference_record['project_id'],
            reference_record['deployment_id'])

        nonblank_annotation_found = False

        im = {}
        im['id'] = image_id
        im['file_name'] = file_name
        im['location'] = location_id
        im['datetime'] = reference_record['timestamp']

        sequence_records_this_sequence = None

        # Should we iterate over image records or sequence records to determine
        # labels for this image?
        label_records = image_records_this_id

        if 'sequence_id' in reference_record:

            assert sequence_id_to_image_ids is not None
            assert 'seq_num_frames' in reference_record, 'sequence processing error'
            assert 'frame_num' in reference_record, 'sequence processing error'

            # Not a typo; WI uses "sequence_id", COCO Camera Traps uses "seq_id"
            im['seq_id'] = reference_record['sequence_id']
            im['seq_num_frames'] = reference_record['seq_num_frames']
            im['frame_num'] = reference_record['frame_num']

            sequence_records_this_sequence = \
                sequence_id_to_sequence_records[reference_record['sequence_id']]
            label_records = sequence_records_this_sequence

            # Image-level and sequence-level taxa should be the same
            #
            # I don't know why labels are reported at both levels.
            taxon_ids_this_sequence = set([r['wi_taxon_id'] for r in sequence_records_this_sequence])
            taxon_ids_each_image = set([r['wi_taxon_id'] for r in image_records_this_id])

            assert taxon_ids_each_image == taxon_ids_this_sequence, \
                'Sequence label inconsistency'

        im['wi_image_info'] = {}
        for s in wi_extra_image_columns:
            assert s in reference_record, \
                'Required column {} missing from image {}'.format(s,reference_record['image_id'])
            im['wi_image_info'][s] = str(reference_record[s])

        categories_this_image = set()

        # Iterate over either image records or label records to determine the labels
        # we should store for this image.
        #
        # record = label_records[0]
        for record in label_records:

            # If there are multiple records for this image (typically because multiple species
            # were recorded), make sure the metadata is consistent across records
            if record != reference_record:

                # "Timestamp" is only present for image records; sequence records use
                # "start_time" and "end_time"
                # assert record['timestamp'] == reference_record['timestamp']
                assert record['project_id'] == reference_record['project_id']
                assert record['deployment_id'] == reference_record['deployment_id']

            count = None

            # This is a bit of future-proofing... it seems odd to me that "count"
            # becomes "number_of_objects" in image-based project downloads.
            if 'count' in record:
                raise ValueError(
                    'Note to self: you suspected a field called "count" might occur in some scenarios')

            # Image-based projects use "number_of_objects"
            if 'number_of_objects' in record:
                assert 'group_size' not in record
                count = record['number_of_objects']

            # Sequence-based projects use "group_size"
            if 'group_size' in record:
                assert 'number_of_objects' not in record
                count = record['group_size']

            if is_empty(count):
                count = None
            else:
                assert is_int(count), \
                    'Illegal group size value: {}'.format(count)
                count = int(count)

            category_name = record['common_name'].strip().lower()

            if len(category_name) == '':

                if len(record['genus']) > 0 and len(record['species']) > 0:
                    category_name = record['genus'] + ' ' + record['species']
                elif len(record['genus']) > 0:
                    category_name = record['genus']
                elif len(record['family']) > 0:
                    category_name = record['family']
                elif len(record['order']) > 0:
                    category_name = record['order']
                elif len(record['class']) > 0:
                    category_name = record['class']
                else:
                    print('Warning: no common name or binomial name available for {}'.format(
                        record['wi_taxon_id']))
                    category_name = record['wi_taxon_id']
                category_name = category_name.strip().lower()

            # ...handling empty category names

            taxonomy_tokens = []
            for level in ('class','order','family','genus','species'):
                taxonomy_tokens.append(record[level])
            taxonomy_string = ';'.join(taxonomy_tokens)
            taxonomy_string = taxonomy_string.lower().strip()

            # Should this category name get remapped?
            if (category_remappings is not None):
                # Check for exact matches
                if category_name in category_remappings:
                    category_name = category_remappings[category_name]
                # Check for regex matches
                else:
                    for k in category_remappings.keys():
                        if re.search(k,category_name):
                            category_name = category_remappings[k]
                            break

            # This is used for logic below, so we handle it outside of category_remappings
            if category_name == 'blank':
                category_name = 'empty'

            assert isinstance(record['is_blank'],int) and \
                record['is_blank'] in (0,1)

            # Resolve disagreements between different ways that blank-ness
            # can be represented
            category_says_blank = category_name == 'empty'
            is_blank_says_blank = record['is_blank'] == 1

            if (category_says_blank) and (is_blank_says_blank):
                category_name = 'empty'
            elif (category_says_blank) or (is_blank_says_blank):
                if blank_disagreement_handling == 'error':
                    raise ValueError('Blank disagreement for {} ({})'.format(
                        image_id, file_name))
                elif blank_disagreement_handling == 'trust_category':
                    print('Warning: category says {}, is_blank says {}, using category'.format(
                        category_name,record['is_blank']))
                elif blank_disagreement_handling == 'trust_is_blank':
                    print('Warning: category says {}, is_blank says {}, using is_blank'.format(
                        category_name,record['is_blank']))
                    if is_blank_says_blank:
                        category_name = 'empty'
                    else:
                        # This is a quirky case, we're supposed to trust is_blank, but
                        # and it says it's not blank, but the category says it is, so we
                        # have no other category we can use
                        assert category_name == 'empty'
                        category_name = 'unknown'

            assert category_name != 'blank'

            # Don't create annotations for the same category twice for the same image
            if category_name in categories_this_image:
                continue
            categories_this_image.add(category_name)

            if category_name in category_name_to_category:
                category = category_name_to_category[category_name]
                category_id = category['id']
                category['count'] = category['count'] + 1
                assert category['name'] == category_name
                if (category_name != 'empty') and \
                   (taxonomy_string != category['taxonomy_string']):
                    print('Warning: category {} has multiple taxonomy strings:\n{}\n{}\n'.format(
                        category_name,
                        taxonomy_string,
                        category['taxonomy_string']))
            else:
                category_id = len(category_name_to_category)
                category = {}
                category_name_to_category[category_name] = category
                category['name'] = category_name
                category['id'] = category_id
                category['count'] = 1
                category['taxonomy_string'] = taxonomy_string

            if category_name != 'empty':
                nonblank_annotation_found = True

            ann = {}
            ann['image_id'] = image_id
            annotations_this_image = image_id_to_annotations[image_id]
            annotation_number = len(annotations_this_image)
            ann['id'] = image_id + '_' + str(annotation_number).zfill(2)
            ann['category_id'] = category_id

            if sequence_records_this_sequence is not None:
                ann['sequence_level_annotation'] = True
            else:
                ann['sequence_level_annotation'] = False

            if count is not None:
                ann['count'] = count

            annotations_this_image.append(ann)

            extra_info = {}
            for s in wi_extra_annotation_columns:
                if s in record:
                    v = record[s]
                    # Don't store empty fields
                    if isinstance(v,str):
                        if (len(v) > 0):
                            extra_info[s] = v
                    # Treat bools as store_true, there are tons of uninformative "False"
                    # fields (e.g. "highlighted").
                    elif isinstance(v,bool):
                        if v:
                            extra_info[s] = v

            if len(extra_info) > 0:
                ann['wi_extra_info'] = extra_info

        # ...for each label record (image or sequence) associated with this image

        if include_blanks or nonblank_annotation_found:
            image_id_to_image[image_id] = im
        else:
            n_blanks_excluded += 1

    # ...for each image


    ##%% Write COCO output

    images = list(image_id_to_image.values())
    categories = list(category_name_to_category.values())

    print('Created COCO records for {} image IDs ({} blanks excluded)'.format(
        len(image_id_to_image),n_blanks_excluded))

    annotations = []

    # image_id_to_annotations contains image IDs we didn't end up using,
    # so we loop over [images] to find the image IDs for which we want to
    # store annotations
    for im in images:
        image_id = im['id']
        annotations_this_image = image_id_to_annotations[image_id]
        for ann in annotations_this_image:
            annotations.append(ann)

    print('Created COCO {} annotation records ({} categories)'.format(
        len(annotations),len(categories)))

    info = {'version':'1.00','description':'converted from WI export'}
    info['source_file'] = csv_file_in
    coco_data = {}
    coco_data['info'] = info
    coco_data['images'] = images
    coco_data['annotations'] = annotations
    coco_data['categories'] = categories

    print_category_counts = False

    if print_category_counts:

        print('Categories and counts:\n')

        category_name_to_count = {c['name']:c['count'] for c in categories}
        category_name_to_count = \
            sort_dictionary_by_value(category_name_to_count,reverse=True)

        for i_category,category_name in enumerate(category_name_to_count):
            category_name_string = category_name
            if (category_name == 'empty') and (not include_blanks):
                category_name_string += (' (excluded)')
            print('{}: {}'.format(category_name_string,
                                category_name_to_count[category_name]))

    ##%% Exclude missing images if requested

    if exclude_missing_images:

        assert os.path.isdir(image_folder), \
            'Must specify a valid image folder if you specify validate_images=True'

        print('Enumerating images in {}'.format(image_folder))
        all_images = find_images(image_folder, return_relative_paths=True, recursive=True)
        all_images_set = set(all_images)

        missing_images = []

        category_name_to_missing_image_count = defaultdict(int)

        category_id_to_name = {c['id']:c['name'] for c in categories}

        # im = images[0]
        for im in tqdm(images):

            file_name_relative = im['file_name']
            if file_name_relative not in all_images_set:

                annotations_this_image = image_id_to_annotations[im['id']]
                categories_this_image = []
                for ann in annotations_this_image:
                    category_id = ann['category_id']
                    category_name = category_id_to_name[category_id]
                    category_name_to_missing_image_count[category_name] += 1

                missing_images.append(im)

        print('Missing {} of {} images'.format(
            len(missing_images),
            len(images)))

        if len(category_name_to_missing_image_count) > 0:

            print('\nCategories with missing images:\n')

            category_name_to_missing_image_count = \
                sort_dictionary_by_value(category_name_to_missing_image_count,
                                         reverse=True)

            for category_name in category_name_to_missing_image_count:
                expected_count_string = ''
                if category_name in category_name_to_count:
                    expected_count_string = ' (of {} in metadata)'.format(
                        category_name_to_count[category_name])
                print('{}: {}{}'.format(category_name,
                                        category_name_to_missing_image_count[category_name],
                                        expected_count_string))

        # ...if we're missing any images

        # TODO: clean up categories that are no longer used
        missing_filenames = set([im['file_name'] for im in missing_images])
        missing_image_ids = set([im['id'] for im in missing_images])

        images = [im for im in images if im['file_name'] not in missing_filenames]
        annotations = [ann for ann in annotations if ann['image_id'] not in missing_image_ids]
        coco_data['images'] = images
        coco_data['annotations'] = annotations

    # ...if we are supposed to exclude missing images


    ##%% Write output json

    if coco_file_out is not None:
        print('Writing COCO data to {}'.format(coco_file_out))
        write_json(coco_file_out,coco_data)


    ##%% Validate output

    from megadetector.data_management.databases.integrity_check_json_db import \
        IntegrityCheckOptions,integrity_check_json_db

    print('Validating COCO file {}'.format(coco_file_out))

    options = IntegrityCheckOptions()
    options.baseDir = image_folder
    options.bCheckImageExistence = False
    options.verbose = verbose

    _ = integrity_check_json_db(coco_file_out,options)

    ##%%

    return coco_data

# ...def wi_download_csv_to_coco(...)


#%% Interactive driver

if False:

    #%%

    image_folder = '/blah/images/2000000'
    csv_file_in = '/csv_downloads/wildlife-insights_046ddddd-d870-dddd-a91d-a50c1a28fe29_project-2001650_data'
    coco_file_out = None
    gs_prefix = 'gs://000000000000_2000000_3658_project_name__main/deployment/'

    validate_images = False
    verbose = True
    category_remappings = default_category_remappings


#%% Command-line driver

# TODO
